import unittest

import numpy as np
import openmdao.api as om
import numpy.testing as npt
import wisdem.commonse.environment as env
from wisdem.commonse import gravity as g
from openmdao.utils.assert_utils import assert_check_partials

npts = 100
myones = np.ones((npts,))


class TestPowerWind(unittest.TestCase):
    def setUp(self):
        self.params = {}
        self.unknowns = {}
        self.resid = None

        self.params["shearExp"] = 2.0
        self.params["Uref"] = 5.0
        self.params["zref"] = 3.0
        self.params["z0"] = 0.0
        self.params["z"] = 9.0 * myones

        self.wind = env.PowerWind(nPoints=npts)

    def testRegular(self):
        self.wind.compute(self.params, self.unknowns)
        expect = 45.0 * myones
        npt.assert_equal(self.unknowns["U"], expect)

    def testIndex(self):
        self.params["z"][1:] = -1.0
        self.wind.compute(self.params, self.unknowns)
        expect = 45.0 * myones
        expect[1:] = 0.0
        npt.assert_equal(self.unknowns["U"], expect)

    def testZ0(self):
        self.params["z0"] = 10.0
        self.params["z"] += 10.0
        self.params["zref"] += 10.0
        self.wind.compute(self.params, self.unknowns)
        expect = 45.0 * myones
        npt.assert_equal(self.unknowns["U"], expect)


class TestLinearWaves(unittest.TestCase):
    def setUp(self):
        self.params = {}
        self.unknowns = {}
        self.resid = None

        self.params["rho_water"] = 1e3
        self.params["Hsig_wave"] = 2.0
        self.params["Uc"] = 5.0
        self.params["z_floor"] = -30.0
        self.params["z_surface"] = 0.0
        self.params["z"] = -2.0 * myones

        self.wave = env.LinearWaves(nPoints=npts)

    def testRegular(self):
        D = np.abs(self.params["z_floor"])
        k = 2.5
        omega = np.sqrt(g * k * np.tanh(k * D))
        self.params["Tsig_wave"] = 2.0 * np.pi / omega

        self.wave.compute(self.params, self.unknowns)
        a = 1.0  # 0.5*Hsig_wave
        z = -2.0
        rho = 1e3
        U_exp = 5 + omega * a * np.cosh(k * (z + D)) / np.sinh(k * D)
        W_exp = -omega * a * np.sinh(k * (z + D)) / np.sinh(k * D)
        V_exp = np.sqrt(U_exp ** 2 + W_exp ** 2)
        A_exp = omega * omega * a * np.cosh(k * (z + D)) / np.sinh(k * D)
        p_exp = -rho * g * (z - a * np.cosh(k * (z + D)) / np.cosh(k * D))

        npt.assert_almost_equal(self.unknowns["U"], U_exp)
        npt.assert_almost_equal(self.unknowns["W"], W_exp)
        npt.assert_almost_equal(self.unknowns["V"], V_exp)
        npt.assert_almost_equal(self.unknowns["A"], A_exp)
        npt.assert_almost_equal(self.unknowns["p"], p_exp)

        # Positive depth input
        self.params["z_floor"] = 30.0
        self.wave.compute(self.params, self.unknowns)
        npt.assert_almost_equal(self.unknowns["U"], U_exp)
        npt.assert_almost_equal(self.unknowns["W"], W_exp)
        npt.assert_almost_equal(self.unknowns["V"], V_exp)
        npt.assert_almost_equal(self.unknowns["A"], A_exp)
        npt.assert_almost_equal(self.unknowns["p"], p_exp)

    def testPositiveZ(self):
        self.params["Tsig_wave"] = 2.0
        self.params["z"] = 2.0 * myones
        self.wave.compute(self.params, self.unknowns)
        npt.assert_equal(self.unknowns["U"], 0.0)
        npt.assert_equal(self.unknowns["W"], 0.0)
        npt.assert_equal(self.unknowns["V"], 0.0)
        npt.assert_equal(self.unknowns["A"], 0.0)
        npt.assert_equal(self.unknowns["p"], 0.0)

    def testQuiet(self):
        self.params["Hsig_wave"] = 0.0
        self.params["Tsig_wave"] = 2.0
        self.wave.compute(self.params, self.unknowns)
        p_exp = 2e3 * g
        npt.assert_equal(self.unknowns["U"], 5.0)
        npt.assert_equal(self.unknowns["W"], 0.0)
        npt.assert_equal(self.unknowns["V"], 5.0)
        npt.assert_equal(self.unknowns["A"], 0.0)
        npt.assert_equal(self.unknowns["p"], p_exp)


class TestPowerWindGradients(unittest.TestCase):
    def test(self):

        z = np.linspace(0.0, 100.0, 20)
        nPoints = len(z)

        prob = om.Problem()
        root = prob.model = om.Group()
        root.add_subsystem("p", env.PowerWind(nPoints=nPoints))

        prob.setup()

        prob["p.Uref"] = 10.0
        prob["p.zref"] = 100.0
        prob["p.z0"] = 0.001  # Fails when z0 = 0, What to do here?
        prob["p.shearExp"] = 0.2

        prob.run_model()

        check = prob.check_partials(out_stream=None, compact_print=True, method="fd")

        assert_check_partials(check)


class TestLogWindGradients(unittest.TestCase):
    def test(self):

        nPoints = 20
        z = np.linspace(0.1, 100.0, nPoints)

        prob = om.Problem()
        root = prob.model = om.Group()
        root.add_subsystem("p", env.LogWind(nPoints=nPoints))

        prob.setup()

        prob["p.Uref"] = 10.0
        prob["p.zref"] = 100.0
        prob["p.z0"] = 0.1  # Fails when z0 = 0

        prob.run_model()

        check = prob.check_partials(out_stream=None, compact_print=True, method="fd")

        assert_check_partials(check)


### These partials are wrong; do not test
# class TestLinearWaveGradients(unittest.TestCase):
#
#     def test(self):
#
#         z_floor = 0.1
#         z_surface = 20.
#         z = np.linspace(z_floor, z_surface, 20)
#         nPoints = len(z)
#
#         prob = om.Problem()
#         root = prob.model = om.Group()
#         root.add_subsystem('p', env.LinearWaves(nPoints=nPoints))
#
#         prob.setup()
#
#         prob['p.Uc'] = 7.0
#         prob['p.z_floor'] = z_floor
#         prob['p.z_surface'] = z_surface
#         prob['p.Hsig_wave'] = 10.0
#         prob['p.Tsig_wave'] = 2.0
#
#         prob.run_model()
#
#         check = prob.check_partials(out_stream=None, compact_print=True, method='fd')
#
#         assert_check_partials(check)


### The partials are currently not correct, so skip this test
# class TestSoilGradients(unittest.TestCase):
#
#     def test(self):
#
#         d0 = 10.0
#         depth = 30.0
#         G = 140e6
#         nu = 0.4
#
#         prob = om.Problem()
#         root = prob.model = om.Group()
#         root.add_subsystem('p', env.TowerSoil())
#
#         prob.setup()
#
#         prob['p.G'] = G
#         prob['p.nu'] = nu
#         prob['p.d0'] = d0
#         prob['p.depth'] = depth
#
#         prob.run_model()
#
#         check = prob.check_partials(out_stream=None, compact_print=True, method='fd')
#
#         assert_check_partials(check)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestPowerWind))
    suite.addTest(unittest.makeSuite(TestLinearWaves))
    suite.addTest(unittest.makeSuite(TestPowerWindGradients))
    suite.addTest(unittest.makeSuite(TestLogWindGradients))
    return suite


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



import numpy as np
import openmdao.api as om
import wisdem.commonse.utilities as util
import wisdem.pyframe3dd.pyframe3dd as pyframe3dd
import wisdem.commonse.utilization_dnvgl as util_dnvgl
import wisdem.commonse.utilization_eurocode as util_euro
import wisdem.commonse.utilization_constraints as util_con
from wisdem.towerse import NFREQ, RIGID, NREFINE, NPTS_SOIL, gravity, get_nfull
from wisdem.commonse.utilization_eurocode import hoopStressEurocode


class TowerPreFrame(om.ExplicitComponent):
    """
    Compute some properties of the tower needed for FEM analysis.

    This component can be simplified by using src_indices for data-passing.
    At the very least, we can code the sparse derivatives as-is for
    input-output relationships.

    Parameters
    ----------
    z_full : numpy array[nFull], [m]
        location along tower. start at bottom and go to top
    suctionpile_depth : float, [m]
        Depth of monopile below sea floor
    transition_piece_mass : float, [kg]
        point mass of transition piece
    gravity_foundation_mass : float, [kg]
        point mass of transition piece
    transition_piece_height : float, [m]
        height of transition piece above water line
    rna_F : numpy array[3], [N]
        rna force
    rna_M : numpy array[3], [N*m]
        rna moment
    k_monopile : numpy array[6], [N/m]
        Stiffness BCs for ocean soil. Only used if monoflag inputis True

    Returns
    -------
    kidx : numpy array[np.int_]
        indices of z where external stiffness reactions should be applied.
    kx : numpy array[nK], [N/m]
        spring stiffness in x-direction
    ky : numpy array[nK], [N/m]
        spring stiffness in y-direction
    kz : numpy array[nK], [N/m]
        spring stiffness in z-direction
    ktx : numpy array[nK], [N/m]
        spring stiffness in theta_x-rotation
    kty : numpy array[nK], [N/m]
        spring stiffness in theta_y-rotation
    ktz : numpy array[nK], [N/m]
        spring stiffness in theta_z-rotation
    midx : numpy array[np.int_]
        indices where added mass should be applied.
    m : numpy array[nMass], [kg]
        added mass
    mIxx : numpy array[nMass], [kg*m**2]
        x mass moment of inertia about some point p
    mIyy : numpy array[nMass], [kg*m**2]
        y mass moment of inertia about some point p
    mIzz : numpy array[nMass], [kg*m**2]
        z mass moment of inertia about some point p
    mIxy : numpy array[nMass], [kg*m**2]
        xy mass moment of inertia about some point p
    mIxz : numpy array[nMass], [kg*m**2]
        xz mass moment of inertia about some point p
    mIyz : numpy array[nMass], [kg*m**2]
        yz mass moment of inertia about some point p
    mrhox : numpy array[nMass], [m]
        x-location of p relative to node
    mrhoy : numpy array[nMass], [m]
        y-location of p relative to node
    mrhoz : numpy array[nMass], [m]
        z-location of p relative to node
    plidx : numpy array[np.int_]
        indices where point loads should be applied.
    Fx : numpy array[nPL], [N]
        point force in x-direction
    Fy : numpy array[nPL], [N]
        point force in y-direction
    Fz : numpy array[nPL], [N]
        point force in z-direction
    Mxx : numpy array[nPL], [N*m]
        point moment about x-axis
    Myy : numpy array[nPL], [N*m]
        point moment about y-axis
    Mzz : numpy array[nPL], [N*m]
        point moment about z-axis

    """

    def initialize(self):
        self.options.declare("n_height")
        self.options.declare("monopile", default=False)
        self.options.declare("soil_springs", default=False)
        self.options.declare("gravity_foundation", default=False)
        self.options.declare("n_refine", default=NREFINE)

    def setup(self):
        n_height = self.options["n_height"]
        n_refine = self.options["n_refine"]
        nFull = get_nfull(n_height, nref=n_refine)

        self.add_input("z_full", np.zeros(nFull), units="m")

        # extra mass
        self.add_input("transition_piece_mass", 0.0, units="kg")
        self.add_input("transition_piece_I", np.zeros(6), units="kg*m**2")
        self.add_input("gravity_foundation_I", np.zeros(6), units="kg*m**2")
        self.add_input("gravity_foundation_mass", 0.0, units="kg")
        self.add_input("transition_piece_height", 0.0, units="m")
        self.add_input("suctionpile_depth", 0.0, units="m")

        # point loads
        self.add_input("rna_F", np.zeros(3), units="N")
        self.add_input("rna_M", np.zeros(3), units="N*m")

        # Monopile handling
        self.add_input("z_soil", np.zeros(NPTS_SOIL), units="N/m")
        self.add_input("k_soil", np.zeros((NPTS_SOIL, 6)), units="N/m")

        # spring reaction data.
        nK = 4 if self.options["monopile"] and not self.options["gravity_foundation"] else 1
        self.add_output("kidx", np.zeros(nK, dtype=np.int_))
        self.add_output("kx", np.zeros(nK), units="N/m")
        self.add_output("ky", np.zeros(nK), units="N/m")
        self.add_output("kz", np.zeros(nK), units="N/m")
        self.add_output("ktx", np.zeros(nK), units="N/m")
        self.add_output("kty", np.zeros(nK), units="N/m")
        self.add_output("ktz", np.zeros(nK), units="N/m")

        # extra mass
        nMass = 2
        self.add_output("midx", np.zeros(nMass, dtype=np.int_))
        self.add_output("m", np.zeros(nMass), units="kg")
        self.add_output("mIxx", np.zeros(nMass), units="kg*m**2")
        self.add_output("mIyy", np.zeros(nMass), units="kg*m**2")
        self.add_output("mIzz", np.zeros(nMass), units="kg*m**2")
        self.add_output("mIxy", np.zeros(nMass), units="kg*m**2")
        self.add_output("mIxz", np.zeros(nMass), units="kg*m**2")
        self.add_output("mIyz", np.zeros(nMass), units="kg*m**2")
        self.add_output("mrhox", np.zeros(nMass), units="m")
        self.add_output("mrhoy", np.zeros(nMass), units="m")
        self.add_output("mrhoz", np.zeros(nMass), units="m")

        # point loads (if addGravityLoadForExtraMass=True be sure not to double count by adding those force here also)
        nPL = 1
        self.add_output("plidx", np.zeros(nPL, dtype=np.int_))
        self.add_output("Fx", np.zeros(nPL), units="N")
        self.add_output("Fy", np.zeros(nPL), units="N")
        self.add_output("Fz", np.zeros(nPL), units="N")
        self.add_output("Mxx", np.zeros(nPL), units="N*m")
        self.add_output("Myy", np.zeros(nPL), units="N*m")
        self.add_output("Mzz", np.zeros(nPL), units="N*m")

        # self.declare_partials("Fx", ["rna_F"], method="fd")
        # self.declare_partials("Fy", ["rna_F"], method="fd")
        # self.declare_partials("Fz", ["rna_F"], method="fd")
        # self.declare_partials("G_full", [], method="fd")
        # self.declare_partials("Mxx", ["rna_M"], method="fd")
        # self.declare_partials("Myy", ["rna_M"], method="fd")
        # self.declare_partials("Mzz", ["rna_M"], method="fd")
        # self.declare_partials("kidx", [], method="fd")
        # self.declare_partials("ktx", [], method="fd")
        # self.declare_partials("kty", [], method="fd")
        # self.declare_partials("ktz", [], method="fd")
        # self.declare_partials("kx", [], method="fd")
        # self.declare_partials("ky", [], method="fd")
        # self.declare_partials("kz", [], method="fd")
        # self.declare_partials("m", ["gravity_foundation_mass", "mass", "transition_piece_mass"], method="fd")
        # self.declare_partials("mIxx", ["gravity_foundation_mass", "mI", "transition_piece_mass"], method="fd")
        # self.declare_partials("mIxy", ["mI"], method="fd")
        # self.declare_partials("mIxz", ["mI"], method="fd")
        # self.declare_partials("mIyy", ["gravity_foundation_mass", "mI", "transition_piece_mass"], method="fd")
        # self.declare_partials("mIyz", ["mI"], method="fd")
        # self.declare_partials("mIzz", ["gravity_foundation_mass", "mI", "transition_piece_mass"], method="fd")
        # self.declare_partials("midx", [], method="fd")
        # self.declare_partials("mrhox", ["mrho"], method="fd")
        # self.declare_partials("mrhoy", ["mrho"], method="fd")
        # self.declare_partials("mrhoz", ["mrho"], method="fd")
        # self.declare_partials("plidx", [], method="fd")

    def compute(self, inputs, outputs):
        n_height = self.options["n_height"]
        n_refine = self.options["n_refine"]
        nFull = get_nfull(n_height, nref=n_refine)
        z = inputs["z_full"]

        # Prepare RNA, transition piece, and gravity foundation (if any applicable) for "extra node mass"
        itrans = util.find_nearest(z, inputs["transition_piece_height"])
        mtrans = float(inputs["transition_piece_mass"])
        Itrans = inputs["transition_piece_I"].flatten()
        mgrav = float(inputs["gravity_foundation_mass"])
        Igrav = inputs["gravity_foundation_I"].flatten()
        # Note, need len()-1 because Frame3DD crashes if mass add at end
        outputs["midx"] = np.array([itrans, 0], dtype=np.int_)
        outputs["m"] = np.array([mtrans, mgrav])
        outputs["mIxx"] = np.array([Itrans[0], Igrav[0]])
        outputs["mIyy"] = np.array([Itrans[1], Igrav[1]])
        outputs["mIzz"] = np.array([Itrans[2], Igrav[2]])
        outputs["mIxy"] = np.array([Itrans[3], Igrav[3]])
        outputs["mIxz"] = np.array([Itrans[4], Igrav[4]])
        outputs["mIyz"] = np.array([Itrans[5], Igrav[5]])
        outputs["mrhox"] = outputs["mrhoy"] = outputs["mrhoz"] = np.zeros(2)

        # Prepare point forces at RNA node
        outputs["plidx"] = np.array([nFull - 1], dtype=np.int_)  # -1 b/c same reason as above
        outputs["Fx"] = np.array([inputs["rna_F"][0]]).flatten()
        outputs["Fy"] = np.array([inputs["rna_F"][1]]).flatten()
        outputs["Fz"] = np.array([inputs["rna_F"][2]]).flatten()
        outputs["Mxx"] = np.array([inputs["rna_M"][0]]).flatten()
        outputs["Myy"] = np.array([inputs["rna_M"][1]]).flatten()
        outputs["Mzz"] = np.array([inputs["rna_M"][2]]).flatten()

        # Prepare for reactions: rigid at tower base
        if self.options["monopile"] and not self.options["gravity_foundation"]:
            if self.options["soil_springs"]:
                z_soil = inputs["z_soil"]
                k_soil = inputs["k_soil"]
                z_pile = z[z <= (z[0] + 1e-1 + np.abs(z_soil[0]))]
                if z_pile.size != 4:
                    print(z)
                    print(z_soil)
                    print(z_pile)
                    raise ValueError("Please use only one section for submerged pile for now")
                k_mono = np.zeros((z_pile.size, 6))
                for k in range(6):
                    k_mono[:, k] = np.interp(z_pile + np.abs(z_soil[0]), z_soil, k_soil[:, k])
                outputs["kidx"] = np.arange(len(z_pile), dtype=np.int_)
                outputs["kx"] = np.array([k_mono[:, 0]])
                outputs["ky"] = np.array([k_mono[:, 2]])
                outputs["kz"] = np.zeros(k_mono.shape[0])
                outputs["kz"][0] = np.array([k_mono[0, 4]])
                outputs["ktx"] = np.array([k_mono[:, 1]])
                outputs["kty"] = np.array([k_mono[:, 3]])
                outputs["ktz"] = np.array([k_mono[:, 5]])

            else:
                z_pile = z[z <= (z[0] + 1e-1 + inputs["suctionpile_depth"])]
                npile = z_pile.size
                if npile != 4:
                    print(z)
                    print(z_pile)
                    print(inputs["suctionpile_depth"])
                    raise ValueError("Please use only one section for submerged pile for now")
                outputs["kidx"] = np.arange(npile, dtype=np.int_)
                outputs["kx"] = outputs["ky"] = outputs["kz"] = RIGID * np.ones(npile)
                outputs["ktx"] = outputs["kty"] = outputs["ktz"] = RIGID * np.ones(npile)

        else:
            outputs["kidx"] = np.array([0], dtype=np.int_)
            outputs["kx"] = outputs["ky"] = outputs["kz"] = np.array([RIGID])
            outputs["ktx"] = outputs["kty"] = outputs["ktz"] = np.array([RIGID])


class CylinderFrame3DD(om.ExplicitComponent):
    """
    Run Frame3DD on the cylinder geometry

    Parameters
    ----------
    z : numpy array[npts], [m]
        location along cylinder. start at bottom and go to top
    Az : numpy array[npts-1], [m**2]
        cross-sectional area
    Asx : numpy array[npts-1], [m**2]
        x shear area
    Asy : numpy array[npts-1], [m**2]
        y shear area
    Jz : numpy array[npts-1], [m**4]
        polar moment of inertia
    Ixx : numpy array[npts-1], [m**4]
        area moment of inertia about x-axis
    Iyy : numpy array[npts-1], [m**4]
        area moment of inertia about y-axis
    E : numpy array[npts-1], [N/m**2]
        modulus of elasticity
    G : numpy array[npts-1], [N/m**2]
        shear modulus
    rho : numpy array[npts-1], [kg/m**3]
        material density
    d : numpy array[npts], [m]
        effective cylinder diameter for section
    t : numpy array[npts-1], [m]
        effective shell thickness for section
    kidx : numpy array[nK, dtype]
        indices of z where external stiffness reactions should be applied.
    kx : numpy array[nK], [N/m]
        spring stiffness in x-direction
    ky : numpy array[nK], [N/m]
        spring stiffness in y-direction
    kz : numpy array[nK], [N/m]
        spring stiffness in z-direction
    ktx : numpy array[nK], [N/m]
        spring stiffness in theta_x-rotation
    kty : numpy array[nK], [N/m]
        spring stiffness in theta_y-rotation
    ktz : numpy array[nK], [N/m]
        spring stiffness in theta_z-rotation
    midx : numpy array[nMass, dtype]
        indices where added mass should be applied.
    m : numpy array[nMass], [kg]
        added mass
    mIxx : numpy array[nMass], [kg*m**2]
        x mass moment of inertia about some point p
    mIyy : numpy array[nMass], [kg*m**2]
        y mass moment of inertia about some point p
    mIzz : numpy array[nMass], [kg*m**2]
        z mass moment of inertia about some point p
    mIxy : numpy array[nMass], [kg*m**2]
        xy mass moment of inertia about some point p
    mIxz : numpy array[nMass], [kg*m**2]
        xz mass moment of inertia about some point p
    mIyz : numpy array[nMass], [kg*m**2]
        yz mass moment of inertia about some point p
    mrhox : numpy array[nMass], [m]
        x-location of p relative to node
    mrhoy : numpy array[nMass], [m]
        y-location of p relative to node
    mrhoz : numpy array[nMass], [m]
        z-location of p relative to node
    plidx : numpy array[nPL, dtype]
        indices where point loads should be applied.
    Fx : numpy array[nPL], [N]
        point force in x-direction
    Fy : numpy array[nPL], [N]
        point force in y-direction
    Fz : numpy array[nPL], [N]
        point force in z-direction
    Mxx : numpy array[nPL], [N*m]
        point moment about x-axis
    Myy : numpy array[nPL], [N*m]
        point moment about y-axis
    Mzz : numpy array[nPL], [N*m]
        point moment about z-axis
    Px : numpy array[nFull], [N/m]
        force per unit length in x-direction
    Py : numpy array[nFull], [N/m]
        force per unit length in y-direction
    Pz : numpy array[nFull], [N/m]
        force per unit length in z-direction

    Returns
    -------
    mass : float, [kg]
        Structural mass computed by Frame3DD
    f1 : float, [Hz]
        First natural frequency
    f2 : float, [Hz]
        Second natural frequency
    structural_frequencies : numpy array[NFREQ], [Hz]
        First and second natural frequency
    fore_aft_freqs : numpy array[NFREQ2]
        Frequencies associated with mode shapes in the tower fore-aft direction
    side_side_freqs : numpy array[NFREQ2]
        Frequencies associated with mode shapes in the tower side-side direction
    torsion_freqs : numpy array[NFREQ2]
        Frequencies associated with mode shapes in the tower torsion direction
    fore_aft_modes : numpy array[NFREQ2, 5]
        6-degree polynomial coefficients of mode shapes in the tower fore-aft direction
        (without constant term)
    side_side_modes : numpy array[NFREQ2, 5]
        6-degree polynomial coefficients of mode shapes in the tower side-side direction
        (without constant term)
    torsion_modes : numpy array[NFREQ2, 5]
        6-degree polynomial coefficients of mode shapes in the tower torsion direction
        (without constant term)
    tower_deflection : numpy array[nFull], [m]
        Deflection of tower nodes in yaw-aligned +x direction
    top_deflection : float, [m]
        Deflection of tower top in yaw-aligned +x direction
    tower_Fz : numpy array[nFull-1], [N]
        Axial foce in vertical z-direction in cylinder structure.
    tower_Vx : numpy array[nFull-1], [N]
        Shear force in x-direction in cylinder structure.
    tower_Vy : numpy array[nFull-1], [N]
        Shear force in y-direction in cylinder structure.
    tower_Mxx : numpy array[nFull-1], [N*m]
        Moment about x-axis in cylinder structure.
    tower_Myy : numpy array[nFull-1], [N*m]
        Moment about y-axis in cylinder structure.
    tower_Mzz : numpy array[nFull-1], [N*m]
        Moment about z-axis in cylinder structure.
    base_F : numpy array[3], [N]
        Total force on cylinder
    base_M : numpy array[3], [N*m]
        Total moment on cylinder measured at base
    """

    def initialize(self):
        self.options.declare("nFull")
        self.options.declare("nK")
        self.options.declare("nMass")
        self.options.declare("nPL")
        self.options.declare("frame3dd_opt")

    def setup(self):
        nFull = self.options["nFull"]
        nK = self.options["nK"]
        nMass = self.options["nMass"]
        nPL = self.options["nPL"]

        # cross-sectional data along cylinder.
        self.add_input("z", val=np.zeros(nFull), units="m")
        self.add_input("Az", val=np.zeros(nFull - 1), units="m**2")
        self.add_input("Asx", val=np.zeros(nFull - 1), units="m**2")
        self.add_input("Asy", val=np.zeros(nFull - 1), units="m**2")
        self.add_input("Jz", val=np.zeros(nFull - 1), units="m**4")
        self.add_input("Ixx", val=np.zeros(nFull - 1), units="m**4")
        self.add_input("Iyy", val=np.zeros(nFull - 1), units="m**4")
        self.add_input("E", val=np.zeros(nFull - 1), units="N/m**2")
        self.add_input("G", val=np.zeros(nFull - 1), units="N/m**2")
        self.add_input("rho", val=np.zeros(nFull - 1), units="kg/m**3")

        # effective geometry -- used for handbook methods to estimate hoop stress, buckling, fatigue
        self.add_input("d", val=np.zeros(nFull), units="m")
        self.add_input("t", val=np.zeros(nFull - 1), units="m")

        # spring reaction data.  Use global RIGID for rigid constraints.
        # JPJ: this next output should be a discrete output
        self.add_input("kidx", val=np.zeros(nK, dtype=np.int_))
        self.add_input("kx", val=np.zeros(nK), units="N/m")
        self.add_input("ky", val=np.zeros(nK), units="N/m")
        self.add_input("kz", val=np.zeros(nK), units="N/m")
        self.add_input("ktx", val=np.zeros(nK), units="N/m")
        self.add_input("kty", val=np.zeros(nK), units="N/m")
        self.add_input("ktz", val=np.zeros(nK), units="N/m")

        # extra mass
        # JPJ: this next output should be a discrete output
        self.add_input("midx", val=np.zeros(nMass, dtype=np.int_))
        self.add_input("m", val=np.zeros(nMass), units="kg")
        self.add_input("mIxx", val=np.zeros(nMass), units="kg*m**2")
        self.add_input("mIyy", val=np.zeros(nMass), units="kg*m**2")
        self.add_input("mIzz", val=np.zeros(nMass), units="kg*m**2")
        self.add_input("mIxy", val=np.zeros(nMass), units="kg*m**2")
        self.add_input("mIxz", val=np.zeros(nMass), units="kg*m**2")
        self.add_input("mIyz", val=np.zeros(nMass), units="kg*m**2")
        self.add_input("mrhox", val=np.zeros(nMass), units="m")
        self.add_input("mrhoy", val=np.zeros(nMass), units="m")
        self.add_input("mrhoz", val=np.zeros(nMass), units="m")

        # point loads (if addGravityLoadForExtraMass=True be sure not to double count by adding those force here also)
        self.add_input("plidx", val=np.zeros(nPL, dtype=np.int_))
        self.add_input("Fx", val=np.zeros(nPL), units="N")
        self.add_input("Fy", val=np.zeros(nPL), units="N")
        self.add_input("Fz", val=np.zeros(nPL), units="N")
        self.add_input("Mxx", val=np.zeros(nPL), units="N*m")
        self.add_input("Myy", val=np.zeros(nPL), units="N*m")
        self.add_input("Mzz", val=np.zeros(nPL), units="N*m")

        # combined wind-water distributed loads
        self.add_input("Px", val=np.zeros(nFull), units="N/m")
        self.add_input("Py", val=np.zeros(nFull), units="N/m")
        self.add_input("Pz", val=np.zeros(nFull), units="N/m")

        # Frequencies
        NFREQ2 = int(NFREQ / 2)
        self.add_output("mass", val=0.0, units="kg")
        self.add_output("f1", val=0.0, units="Hz")
        self.add_output("f2", val=0.0, units="Hz")
        self.add_output("structural_frequencies", np.zeros(NFREQ), units="Hz")
        self.add_output("fore_aft_modes", np.zeros((NFREQ2, 5)))
        self.add_output("side_side_modes", np.zeros((NFREQ2, 5)))
        self.add_output("torsion_modes", np.zeros((NFREQ2, 5)))
        self.add_output("fore_aft_freqs", np.zeros(NFREQ2), units="Hz")
        self.add_output("side_side_freqs", np.zeros(NFREQ2), units="Hz")
        self.add_output("torsion_freqs", np.zeros(NFREQ2), units="Hz")
        self.add_output("tower_deflection", np.zeros(nFull), units="m")
        self.add_output("top_deflection", 0.0, units="m")
        self.add_output("tower_Fz", val=np.zeros(nFull - 1), units="N")
        self.add_output("tower_Vx", val=np.zeros(nFull - 1), units="N")
        self.add_output("tower_Vy", val=np.zeros(nFull - 1), units="N")
        self.add_output("tower_Mxx", val=np.zeros(nFull - 1), units="N*m")
        self.add_output("tower_Myy", val=np.zeros(nFull - 1), units="N*m")
        self.add_output("tower_Mzz", val=np.zeros(nFull - 1), units="N*m")
        self.add_output("base_F", val=np.zeros(3), units="N")
        self.add_output("base_M", val=np.zeros(3), units="N*m")

        # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

    def compute(self, inputs, outputs):

        frame3dd_opt = self.options["frame3dd_opt"]

        # ------- node data ----------------
        z = inputs["z"]
        n = len(z)
        node = np.arange(1, n + 1)
        x = np.zeros(n)
        y = np.zeros(n)
        r = np.zeros(n)

        nodes = pyframe3dd.NodeData(node, x, y, z, r)
        # -----------------------------------

        # ------ reaction data ------------

        # rigid base
        node = inputs["kidx"] + np.ones(len(inputs["kidx"]))  # add one because 0-based index but 1-based node numbering
        rigid = RIGID

        reactions = pyframe3dd.ReactionData(
            node, inputs["kx"], inputs["ky"], inputs["kz"], inputs["ktx"], inputs["kty"], inputs["ktz"], rigid
        )
        # -----------------------------------

        # ------ frame element data ------------
        element = np.arange(1, n)
        N1 = np.arange(1, n)
        N2 = np.arange(2, n + 1)

        roll = np.zeros(n - 1)

        # Element properties
        Az = inputs["Az"]
        Asx = inputs["Asx"]
        Asy = inputs["Asy"]
        Jz = inputs["Jz"]
        Ixx = inputs["Ixx"]
        Iyy = inputs["Iyy"]
        E = inputs["E"]
        G = inputs["G"]
        rho = inputs["rho"]

        elements = pyframe3dd.ElementData(element, N1, N2, Az, Asx, Asy, Jz, Ixx, Iyy, E, G, roll, rho)
        # -----------------------------------

        # ------ options ------------
        dx = -1.0
        options = pyframe3dd.Options(frame3dd_opt["shear"], frame3dd_opt["geom"], dx)
        # -----------------------------------

        # initialize frame3dd object
        cylinder = pyframe3dd.Frame(nodes, reactions, elements, options)

        # ------ add extra mass ------------

        # extra node inertia data
        N = inputs["midx"] + np.ones(len(inputs["midx"]))

        add_gravity = True
        cylinder.changeExtraNodeMass(
            N,
            inputs["m"],
            inputs["mIxx"],
            inputs["mIyy"],
            inputs["mIzz"],
            inputs["mIxy"],
            inputs["mIxz"],
            inputs["mIyz"],
            inputs["mrhox"],
            inputs["mrhoy"],
            inputs["mrhoz"],
            add_gravity,
        )

        # ------------------------------------

        # ------- enable dynamic analysis ----------
        Mmethod = 1
        lump = 0
        shift = 0.0
        # Run twice the number of modes to ensure that we can ignore the torsional modes and still get the desired number of fore-aft, side-side modes
        cylinder.enableDynamics(2 * NFREQ, Mmethod, lump, frame3dd_opt["tol"], shift)
        # ----------------------------

        # ------ static load case 1 ------------

        # gravity in the X, Y, Z, directions (global)
        gx = 0.0
        gy = 0.0
        gz = -gravity

        load = pyframe3dd.StaticLoadCase(gx, gy, gz)

        # point loads
        nF = inputs["plidx"] + np.ones(len(inputs["plidx"]))
        load.changePointLoads(nF, inputs["Fx"], inputs["Fy"], inputs["Fz"], inputs["Mxx"], inputs["Myy"], inputs["Mzz"])

        # distributed loads
        Px, Py, Pz = inputs["Pz"], inputs["Py"], -inputs["Px"]  # switch to local c.s.
        z = inputs["z"]

        # trapezoidally distributed loads
        EL = np.arange(1, n)
        xx1 = xy1 = xz1 = np.zeros(n - 1)
        xx2 = xy2 = xz2 = np.diff(z) - 1e-6  # subtract small number b.c. of precision
        wx1 = Px[:-1]
        wx2 = Px[1:]
        wy1 = Py[:-1]
        wy2 = Py[1:]
        wz1 = Pz[:-1]
        wz2 = Pz[1:]

        load.changeTrapezoidalLoads(EL, xx1, xx2, wx1, wx2, xy1, xy2, wy1, wy2, xz1, xz2, wz1, wz2)

        cylinder.addLoadCase(load)
        # Debugging
        # cylinder.write('temp.3dd')
        # -----------------------------------
        # run the analysis
        displacements, forces, reactions, internalForces, mass, modal = cylinder.run()
        ic = 0

        # mass
        outputs["mass"] = mass.struct_mass

        # natural frequncies
        outputs["f1"] = modal.freq[0]
        outputs["f2"] = modal.freq[1]
        outputs["structural_frequencies"] = modal.freq[:NFREQ]

        # Get all mode shapes in batch
        NFREQ2 = int(NFREQ / 2)
        freq_x, freq_y, freq_z, mshapes_x, mshapes_y, mshapes_z = util.get_xyz_mode_shapes(
            z, modal.freq, modal.xdsp, modal.ydsp, modal.zdsp, modal.xmpf, modal.ympf, modal.zmpf
        )
        outputs["fore_aft_freqs"] = freq_x[:NFREQ2]
        outputs["side_side_freqs"] = freq_y[:NFREQ2]
        outputs["torsion_freqs"] = freq_z[:NFREQ2]
        outputs["fore_aft_modes"] = mshapes_x[:NFREQ2, :]
        outputs["side_side_modes"] = mshapes_y[:NFREQ2, :]
        outputs["torsion_modes"] = mshapes_z[:NFREQ2, :]

        # deflections due to loading (from cylinder top and wind/wave loads)
        outputs["tower_deflection"] = np.sqrt(
            displacements.dx[ic, :] ** 2 + displacements.dy[ic, :] ** 2
        )  # in yaw-aligned direction
        outputs["top_deflection"] = outputs["tower_deflection"][-1]

        # Record total forces and moments
        ibase = 2 * int(inputs["kidx"].max())
        outputs["base_F"] = -np.r_[-forces.Vz[ic, ibase], forces.Vy[ic, ibase], forces.Nx[ic, ibase]]
        outputs["base_M"] = -np.r_[-forces.Mzz[ic, ibase], forces.Myy[ic, ibase], forces.Txx[ic, ibase]]

        # Forces and moments along the structure
        outputs["tower_Fz"] = forces.Nx[ic, 1::2]
        outputs["tower_Vx"] = -forces.Vz[ic, 1::2]
        outputs["tower_Vy"] = forces.Vy[ic, 1::2]
        outputs["tower_Mxx"] = -forces.Mzz[ic, 1::2]
        outputs["tower_Myy"] = forces.Myy[ic, 1::2]
        outputs["tower_Mzz"] = forces.Txx[ic, 1::2]


class TowerPostFrame(om.ExplicitComponent):
    """
    Postprocess results from Frame3DD.

    Parameters
    ----------
    z_full : numpy array[nFull], [m]
        location along tower. start at bottom and go to top
    d_full : numpy array[nFull], [m]
        effective tower diameter for section
    t_full : numpy array[nFull-1], [m]
        effective shell thickness for section
    suctionpile_depth : float, [m]
        Depth of monopile below sea floor
    E_full : numpy array[nFull-1], [Pa]
        Isotropic Youngs modulus of the materials along the tower sections.
    G_full : numpy array[nFull-1], [Pa]
        Isotropic shear modulus of the materials along the tower sections.
    rho_full : numpy array[nFull-1], [kg/m**3]
        Density of the materials along the tower sections.
    sigma_y_full : numpy array[nFull-1], [N/m**2]
        yield stress
    tower_Fz : numpy array[nFull-1], [N]
        Axial foce in vertical z-direction in cylinder structure.
    tower_Vx : numpy array[nFull-1], [N]
        Shear force in x-direction in cylinder structure.
    tower_Vy : numpy array[nFull-1], [N]
        Shear force in y-direction in cylinder structure.
    tower_Mxx : numpy array[nFull-1], [N*m]
        Moment about x-axis in cylinder structure.
    tower_Myy : numpy array[nFull-1], [N*m]
        Moment about y-axis in cylinder structure.
    tower_Mzz : numpy array[nFull-1], [N*m]
        Moment about z-axis in cylinder structure.
    qdyn : numpy array[nFull], [N/m**2]
        dynamic pressure

    Returns
    -------
    axial_stress : numpy array[nFull-1], [N/m**2]
        Axial stress in cylinder structure
    shear_stress : numpy array[nFull-1], [N/m**2]
        Shear stress in cylinder structure
    hoop_stress : numpy array[nFull-1], [N/m**2]
        Hoop stress in cylinder structure calculated with simple method used in API
        standards
    hoop_stress_euro : numpy array[nFull-1], [N/m**2]
        Hoop stress in cylinder structure calculated with Eurocode method
    stress : numpy array[nFull-1]
        Von Mises stress utilization along tower at specified locations. Includes safety
        factor.
    shell_buckling : numpy array[nFull-1]
        Shell buckling constraint. Should be < 1 for feasibility. Includes safety
        factors
    global_buckling : numpy array[nFull-1]
        Global buckling constraint. Should be < 1 for feasibility. Includes safety
        factors
    turbine_F : numpy array[3], [N]
        Total force on tower+rna
    turbine_M : numpy array[3], [N*m]
        Total x-moment on tower+rna measured at base

    """

    def initialize(self):
        self.options.declare("modeling_options")
        # self.options.declare('nDEL')

    def setup(self):
        n_height = self.options["modeling_options"]["n_height"]
        n_refine = self.options["modeling_options"]["n_refine"]
        nFull = get_nfull(n_height, nref=n_refine)

        # effective geometry -- used for handbook methods to estimate hoop stress, buckling, fatigue
        self.add_input("z_full", np.zeros(nFull), units="m")
        self.add_input("d_full", np.zeros(nFull), units="m")
        self.add_input("t_full", np.zeros(nFull - 1), units="m")
        self.add_input("suctionpile_depth", 0.0, units="m")

        self.add_input("Az", val=np.zeros(nFull - 1), units="m**2")
        self.add_input("Asx", val=np.zeros(nFull - 1), units="m**2")
        self.add_input("Asy", val=np.zeros(nFull - 1), units="m**2")
        self.add_input("Jz", val=np.zeros(nFull - 1), units="m**4")
        self.add_input("Ixx", val=np.zeros(nFull - 1), units="m**4")
        self.add_input("Iyy", val=np.zeros(nFull - 1), units="m**4")

        # Material properties
        self.add_input("E_full", np.zeros(nFull - 1), units="N/m**2")
        self.add_input("G_full", np.zeros(nFull - 1), units="Pa")
        self.add_input("rho_full", np.zeros(nFull - 1), units="kg/m**3")
        self.add_input("sigma_y_full", np.zeros(nFull - 1), units="N/m**2")

        # Processed Frame3DD/OpenFAST outputs
        self.add_input("tower_Fz", val=np.zeros(nFull - 1), units="N")
        self.add_input("tower_Vx", val=np.zeros(nFull - 1), units="N")
        self.add_input("tower_Vy", val=np.zeros(nFull - 1), units="N")
        self.add_input("tower_Mxx", val=np.zeros(nFull - 1), units="N*m")
        self.add_input("tower_Myy", val=np.zeros(nFull - 1), units="N*m")
        self.add_input("tower_Mzz", val=np.zeros(nFull - 1), units="N*m")
        self.add_input("qdyn", val=np.zeros(nFull), units="N/m**2")

        # fatigue parameters
        self.add_input("life", 20.0)
        # self.add_input('m_SN', 4, desc='slope of S/N curve')
        # self.add_input('DC', 80.0, desc='standard value of stress')
        # self.add_input('z_DEL', np.zeros(nDEL), units='m', desc='absolute z coordinates of corresponding fatigue parameters')
        # self.add_input('M_DEL', np.zeros(nDEL), desc='fatigue parameters at corresponding z coordinates')

        # Load analysis
        self.add_output("axial_stress", val=np.zeros(nFull - 1), units="N/m**2")
        self.add_output("shear_stress", val=np.zeros(nFull - 1), units="N/m**2")
        self.add_output("hoop_stress", val=np.zeros(nFull - 1), units="N/m**2")

        self.add_output("hoop_stress_euro", val=np.zeros(nFull - 1), units="N/m**2")
        self.add_output("constr_stress", np.zeros(nFull - 1))
        self.add_output("constr_shell_buckling", np.zeros(nFull - 1))
        self.add_output("constr_global_buckling", np.zeros(nFull - 1))
        # self.add_output('constr_damage', np.zeros(nFull-1), desc='Fatigue damage at each tower section')
        self.add_output("turbine_F", val=np.zeros(3), units="N", desc="Total force on tower+rna")
        self.add_output("turbine_M", val=np.zeros(3), units="N*m", desc="Total x-moment on tower+rna measured at base")

    def compute(self, inputs, outputs):
        # Unpack some variables
        sigma_y = inputs["sigma_y_full"]
        E = inputs["E_full"]
        G = inputs["G_full"]
        z = inputs["z_full"]
        t = inputs["t_full"]
        d = inputs["d_full"]
        h = np.diff(z)
        d_sec, _ = util.nodal2sectional(d)
        r_sec = 0.5 * d_sec
        n_sec = r_sec.size

        L_suction = float(inputs["suctionpile_depth"])
        L_buckling = self.options["modeling_options"]["buckling_length"]
        gamma_f = self.options["modeling_options"]["gamma_f"]
        gamma_m = self.options["modeling_options"]["gamma_m"]
        gamma_n = self.options["modeling_options"]["gamma_n"]
        gamma_b = self.options["modeling_options"]["gamma_b"]

        # axial and shear stress
        qdyn, _ = util.nodal2sectional(inputs["qdyn"])

        ##R = self.d/2.0
        ##x_stress = R*np.cos(self.theta_stress)
        ##y_stress = R*np.sin(self.theta_stress)
        ##axial_stress = Fz/self.Az + Mxx/self.Ixx*y_stress - Myy/Iyy*x_stress
        #        V = Vy*x_stress/R - Vx*y_stress/R  # shear stress orthogonal to direction x,y
        #        shear_stress = 2. * V / self.Az  # coefficient of 2 for a hollow circular section, but should be conservative for other shapes

        # Get loads from Framee3dd/OpenFAST
        Fz = inputs["tower_Fz"]
        Vx = inputs["tower_Vx"]
        Vy = inputs["tower_Vy"]
        Mxx = inputs["tower_Mxx"]
        Myy = inputs["tower_Myy"]
        Mzz = inputs["tower_Mzz"]

        M = np.sqrt(Mxx ** 2 + Myy ** 2)
        V = np.sqrt(Vx ** 2 + Vy ** 2)

        # Geom properties
        Az = inputs["Az"]
        Asx = inputs["Asx"]
        Asy = inputs["Asy"]
        Jz = inputs["Jz"]
        Ixx = inputs["Ixx"]
        Iyy = inputs["Iyy"]

        # See http://svn.code.sourceforge.net/p/frame3dd/code/trunk/doc/Frame3DD-manual.html#structuralmodeling
        outputs["axial_stress"] = axial_stress = Fz / Az + M * r_sec / Iyy
        outputs["shear_stress"] = shear_stress = np.abs(Mzz) / Jz * r_sec + V / Asx
        outputs["hoop_stress"] = hoop_stress = util_con.hoopStress(d_sec, t, qdyn)
        outputs["constr_stress"] = util_con.vonMisesStressUtilization(
            axial_stress, hoop_stress, shear_stress, gamma_f * gamma_m * gamma_n, sigma_y
        )

        if self.options["modeling_options"]["buckling_method"].lower().find("euro") >= 0:
            # Use Euro-code method
            L_buckling = L_buckling * np.ones(axial_stress.shape)
            hoop_euro = util_euro.hoopStressEurocode(d_sec, t, L_buckling, hoop_stress)
            outputs["hoop_stress_euro"] = hoop_euro

            shell_buckling = util_euro.shellBucklingEurocode(
                d, t, axial_stress, hoop_euro, shear_stress, L_buckling, E, sigma_y, gamma_f, gamma_b
            )

            tower_height = z[-1] - z[0] - L_suction
            global_buckling = util_euro.bucklingGL(d_sec, t, Fz, M, tower_height, E, sigma_y, gamma_f, gamma_b)

        else:
            # Use DNV-GL CP202 Method
            check = util_dnvgl.CylinderBuckling(
                h, d, t, E=E, G=G, sigma_y=sigma_y, gamma=gamma_f * gamma_b, mod_length=L_suction
            )
            results = check.run_buckling_checks(Fz, M, axial_stress, hoop_stress, shear_stress)
            shell_buckling = results["Shell"]
            global_buckling = results["Global"]

        outputs["constr_shell_buckling"] = shell_buckling
        outputs["constr_global_buckling"] = global_buckling

        # fatigue
        # N_DEL = 365.0 * 24.0 * 3600.0 * inputs["life"] * np.ones(len(t))
        # outputs['damage'] = np.zeros(N_DEL.shape)

        # if any(inputs['M_DEL']):
        #    M_DEL = np.interp(z_section, inputs['z_DEL'], inputs['M_DEL'])

        #    outputs['damage'] = util_con.fatigue(M_DEL, N_DEL, d, inputs['t'], inputs['m_SN'],
        #                                      inputs['DC'], gamma_fatigue, stress_factor=1.0, weld_factor=True)



import numpy as np
import pandas as pd


class WeatherDelay:
    """
    Calculates weather delays for a project based on weather data, season of
    construction, operational hours, mission time,
    and critical wind speed -- only calculates wind delays right now.

    Central to this calculation is a weather window, which is a list of hours
    during the times when construction can happen. Each row lists
    an year, month, day and hour with the various weather conditions. Below,
    you can specify a start delay, which is the number of rows in the data--
    or the number of hours at the start of the weather window--during which the
    work will not start.

    The INPUT keys are the following:

    weather_window
        (pd.DataFrame) The weather window as prepared by the
        read_weather_window function in the WeatherWindowCSVReader module.
        See the documentation for that function for details on the
        columns of the weather window.

    start_delay_hours
        (float) Delay of mission from start of weather window. The weather
        window is a list of hours with the weather values as rows. The delay
        can be as little as 0 hours or the total available construction hours
        for the project.

    mission_time_hours
        (float) Length of mission. Mission length is the time it takes to
        complete the operation that may be delayed by the weather.

    critical_wind_speed_m_per_s
        (float) Wind speed that the mission must shutdown and enter a delay state.


    wind_height_of_interest_m
        (float) Height used in wind shear calculations.

    The OUTPUT keys are the following

    wind_delay
        (list) List of number of hours of each wind delay during the mission.
        length of list is number of weather delays. Value in list is duration
        of weather delay in hours.

    Parmeters
    ---------
    input_dict : dict
        The input data

    output_dict : dict
        The output data
    """

    def __init__(self, input_dict, output_dict):
        self.input_dict = input_dict
        self.output_dict = output_dict
        self.validate_inputs(self.input_dict)
        self.output_dict["wind_delays"] = self.calculate_wind_delay()

    def validate_inputs(self, input_dict):
        """
        This method checks a dictionary to make sure it has keys for all
        necessary values needed for calculations in an instance of this
        class. It is made to validate input_dict dictionaries.

        Returns
        -------
        None
            This method returns nothing if the validation passes. If it does
            not pass an exception is raised.

        Raises
        ------
        ValueError
            If one of the keys is missing, this method raises a ValueError
        """
        required_keys = {
            "start_delay_hours",
            "mission_time_hours",
            "critical_wind_speed_m_per_s",
            "wind_height_of_interest_m",
            "wind_shear_exponent",
            "weather_window",
        }
        found_keys = set(input_dict.keys())
        if len(required_keys - found_keys) > 0:
            err_msg = "{}: did not find all required keys in inputs dictionary. Missing keys are {}"
            raise ValueError(err_msg.format(type(self).__name__, required_keys - found_keys))

    def calculate_wind_delay(self):
        """
        Calculates wind delay based on weather window, mission time, and critical wind speed.

        This method must be executed after the create_weather_window method which calculates
        the weather window. This method assumes that self.output['weather_window'] is set.
        That is where it gets its weather window.

        Calculate wind delay based on weather window, operation start
        delay, mission time, and critical wind speed

        Returns
        -------
        list
            Number of hours for each wind delay encountered during mission.
            count of list = number of weather delays.
            values in list = durations of weather delays.
        """

        # Pull function variables off of the dictionaries to make the
        # following lines shorter. Also, the keys on the input
        # dictionary have units but the lines below do not have units in
        # the variable names.
        start_delay = self.input_dict["start_delay_hours"]
        mission_time = self.input_dict["mission_time_hours"]
        critical_wind_speed = self.input_dict["critical_wind_speed_m_per_s"]
        wind_height_of_interest_m = self.input_dict["wind_height_of_interest_m"]
        wind_shear_exponent = self.input_dict["wind_shear_exponent"]
        weather_window = self.input_dict["weather_window"]

        # Extract only the 'Speed m per s' as a dataframe, and only retain
        # elements where index is > start_delay and < mission_time
        wind_speeds_m_s = weather_window["Speed m per s"].values
        # check if mission time exceeds size of weather window
        if mission_time > len(wind_speeds_m_s):
            raise ValueError("{}: Error: Mission time longer than weather window".format(type(self).__name__))
        wind_speeds_m_s_filtered = wind_speeds_m_s[(start_delay + 1) : (int(mission_time) + 1)]

        # Calculate the wind speed at the particular, given the wind shear exponent
        wind_speed_at_height_m_s = wind_speeds_m_s_filtered * (wind_height_of_interest_m / 100) ** wind_shear_exponent

        # wind_delays is an array of booleans. It is True if the critical
        # wind speed is exceeded. False if the critical wind speed is not
        # exceeded. Each element represents an hour of wind
        wind_delays = wind_speed_at_height_m_s > critical_wind_speed

        # If there are any wind delays found:
        #
        # Iterate over each of the wind_delays. Count for contiguous blocks of
        # hours of wind delays. Add the length of each of these blocks to the
        # delay_durations.
        #
        # This code snippet takes O(n) linear time. The trick in vectorizing it
        # is that contiguous blocks of unknown length of True wind delays
        # are needed to find durations. But linear time may not be a problem
        # because the data we are iterating over has already been filtered down
        # to a small set.

        if np.any(wind_delays):

            # Holds the list of delay durations
            delay_durations = []

            # As we go through each delay, this accumulates the duration of
            # the current delay. Each True element of wind_delays increments
            # this by one
            current_delay_duration = 0

            # The following variable is True if we are iterating through
            # that are True, which means a weather delay.
            iterating_through_wind_delay = False

            # Iterate over each element of wind_delays. This is O(n) as
            # noted above.
            for wind_delay in np.nditer(wind_delays):
                # If we are in a weather delay...
                if wind_delay:
                    # If we are not currently iterating over a weather delay
                    # a new continuous sequence of delayed hours has started.
                    if not iterating_through_wind_delay:
                        current_delay_duration = 1
                        iterating_through_wind_delay = True

                    # While iterating over delay, increment counter.
                    else:
                        current_delay_duration += 1

                # If we are NOT iterating through a wind delay
                # And we were iterating through a wind delay, end that
                # delay and record the duration.
                elif iterating_through_wind_delay:
                    delay_durations.append(current_delay_duration)
                    iterating_through_wind_delay = False

                # Otherwise we are not iterating through a wind delay and
                # and did not finish a wind delay, so we do nothing with
                # no need for an else.

            # Finally return the durations we found.
            return delay_durations

        # If there are not wind delays, return a list with just 0 in it
        else:
            return [0]

    def run_module(self):
        """
        This method runs all other methods in the module in order to set the
        correct output keys.

        Returns
        -------
        int
            0 if the module ran without errors. 1 if there was an error.
        """
        try:
            self.output_dict["wind_delay"] = self.calculate_wind_delay()
            return 0  # module ran successfully
        except:
            return 1  # module did not run successfully



import openmdao.api as om


class PlantFinance(om.ExplicitComponent):
    """
    Compute LCOE for the wind plant

    Parameters
    ----------
    machine_rating : float
        Rating of the turbine
    tcc_per_kW : float
        Turbine capital cost
    offset_tcc_per_kW : float
        Offset to turbine capital cost
    turbine_number : int
        Number of turbines at plant
    bos_per_kW : float
        Balance of system costs of the turbine
    opex_per_kW : float
        Average annual operational expenditures of the turbine
    plant_aep_in : float
        Annual Energy Production of the wind plant
    turbine_aep : float
        Annual Energy Production of the wind turbine
    wake_loss_factor : float
        The losses in AEP due to waked conditions
    fixed_charge_rate : float
        Fixed charge rate for coe calculation

    Returns
    -------
    lcoe : float
        Levelized cost of energy for the wind plant
    plant_aep : float
        Annual Energy Production of the wind plant

    """

    def initialize(self):
        self.options.declare("verbosity", default=False)

    def setup(self):

        # Inputs
        self.add_input("machine_rating", val=0.0, units="kW")
        self.add_input("tcc_per_kW", val=0.0, units="USD/kW")
        self.add_input("offset_tcc_per_kW", val=0.0, units="USD/kW")
        self.add_discrete_input("turbine_number", val=0)
        self.add_input("bos_per_kW", val=0.0, units="USD/kW")
        self.add_input("opex_per_kW", val=0.0, units="USD/kW/yr")
        self.add_input("plant_aep_in", val=0.0, units="kW*h")
        self.add_input("turbine_aep", val=0.0, units="kW*h")
        self.add_input("wake_loss_factor", val=0.15)
        self.add_input("fixed_charge_rate", val=0.075)

        self.add_output("lcoe", val=0.0, units="USD/kW/h")
        self.add_output("plant_aep", val=0.0, units="USD/kW/h")

        self.declare_partials("*", "*")

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        # Unpack parameters
        t_rating = inputs["machine_rating"]
        n_turbine = discrete_inputs["turbine_number"]
        tcc_per_kW = inputs["tcc_per_kW"] + inputs["offset_tcc_per_kW"]
        bos_per_kW = inputs["bos_per_kW"]
        opex_per_kW = inputs["opex_per_kW"]
        fcr = inputs["fixed_charge_rate"]
        wlf = inputs["wake_loss_factor"]
        turb_aep = inputs["turbine_aep"]
        c_turbine = tcc_per_kW * t_rating
        c_bos_turbine = bos_per_kW * t_rating
        c_opex_turbine = opex_per_kW * t_rating

        # Run a few checks on the inputs
        if n_turbine == 0:
            raise ValueError(
                "ERROR: The number of the turbines in the plant is not initialized correctly and it is currently equal to 0. Check the connections to Plant_FinanceSE"
            )

        if c_turbine == 0:
            raise ValueError(
                "ERROR: The cost of the turbines in the plant is not initialized correctly and it is currently equal to 0 USD. Check the connections to Plant_FinanceSE"
            )

        if c_bos_turbine == 0:
            print(
                "WARNING: The BoS costs of the turbine are not initialized correctly and they are currently equal to 0 USD. Check the connections to Plant_FinanceSE"
            )

        if c_opex_turbine == 0:
            print(
                "WARNING: The Opex costs of the turbine are not initialized correctly and they are currently equal to 0 USD. Check the connections to Plant_FinanceSE"
            )

        if inputs["plant_aep_in"] == 0:
            if turb_aep == 0:
                print(
                    "WARNING: AEP is not computed properly or it is not connected to PlantFinanceSE. Both turbine_aep and park_aep are currently equal to 0 Wh. Check the connections to Plant_FinanceSE"
                )
            park_aep = n_turbine * turb_aep * (1.0 - wlf)
            dpark_dtaep = n_turbine * (1.0 - wlf)
            dpark_dnturb = turb_aep * (1.0 - wlf)
            dpark_dwlf = -n_turbine * turb_aep
            dpark_dpaep = 0.0
        else:
            park_aep = inputs["plant_aep_in"]
            dpark_dpaep = 1.0
            dpark_dtaep = dpark_dnturb = dpark_dwlf = 0.0

        npr = n_turbine * t_rating  # net park rating, used in net energy capture calculation below
        dnpr_dnturb = t_rating
        dnpr_dtrating = n_turbine

        nec = park_aep / npr  # net energy rating, per COE report
        dnec_dwlf = dpark_dwlf / npr
        dnec_dtaep = dpark_dtaep / npr
        dnec_dpaep = dpark_dpaep / npr
        dnec_dnturb = dpark_dnturb / npr - dnpr_dnturb * nec / npr
        dnec_dtrating = -dnpr_dtrating * nec / npr

        icc = (c_turbine + c_bos_turbine) / t_rating  # $/kW, changed per COE report
        c_opex = (c_opex_turbine) / t_rating  # $/kW, changed per COE report

        dicc_dtrating = -icc / t_rating
        dcopex_dtrating = -c_opex / t_rating
        dicc_dcturb = dicc_dcbos = dcopex_dcopex = 1.0 / t_rating

        # compute COE and LCOE values
        lcoe = (icc * fcr + c_opex) / nec  # changed per COE report
        outputs["lcoe"] = lcoe
        outputs["plant_aep"] = park_aep

        self.J = {}
        self.J["lcoe", "tcc_per_kW"] = dicc_dcturb * fcr / nec
        self.J["lcoe", "turbine_number"] = -dnec_dnturb * lcoe / nec
        self.J["lcoe", "bos_per_kW"] = dicc_dcbos * fcr / nec
        self.J["lcoe", "opex_per_kW"] = dcopex_dcopex / nec
        self.J["lcoe", "fixed_charge_rate"] = icc / nec
        self.J["lcoe", "wake_loss_factor"] = -dnec_dwlf * lcoe / nec
        self.J["lcoe", "turbine_aep"] = -dnec_dtaep * lcoe / nec
        self.J["lcoe", "plant_aep_in"] = -dnec_dpaep * lcoe / nec
        self.J["lcoe", "machine_rating"] = (dicc_dtrating * fcr + dcopex_dtrating) / nec - dnec_dtrating * lcoe / nec

        if self.options["verbosity"]:
            print("################################################")
            print("Computation of LCoE from Plant_FinanceSE")
            print("Number of turbines in the park    %u" % n_turbine)
            print("Turbine rating                    %.2f kW" % t_rating)
            print("Turbine capital cost per kW       %.2f USD/kW" % tcc_per_kW)
            print("BoS costs per kW                  %.2f USD/kW" % bos_per_kW)
            print("Opex costs per kW                 %.2f USD/kW" % opex_per_kW)
            print("Fixed charge rate                 %.2f %%" % (fcr * 100.0))
            print("Wake loss factor                  %.2f %%" % (wlf * 100.0))
            print("AEP of the single turbine         %.2f GWh" % (turb_aep * 1.0e-006))
            print("AEP of the wind plant             %.2f GWh" % (park_aep * 1.0e-006))
            print("Initial capital costs per kW      %.2f $/kW" % icc)
            print("Total initial capital cost        %.2f M USD" % (icc * n_turbine * t_rating * 1.0e-006))
            print("Opex costs of the park            %.2f M USD/yr" % (c_opex_turbine * n_turbine * 1.0e-006))
            print("Net energy capture                %.2f MWh/MW/yr" % nec)
            print(
                "LCoE                              %.2f USD/MWh" % (lcoe * 1.0e003)
            )  # removed "coe", best to have only one metric for cost
            print("################################################")

    def compute_partials(self, inputs, J, discrete_inputs):
        J = self.J



import math
import traceback

import numpy as np
import pandas as pd
from scipy.optimize import root_scalar
from wisdem.landbosse.model.CostModule import CostModule
from wisdem.landbosse.model.WeatherDelay import WeatherDelay as WD


class FoundationCost(CostModule):
    """
    **FoundationCost.py**

    - Created by Annika Eberle and Owen Roberts on Apr. 3, 2018

    - Refactored by Parangat Bhaskar and Alicia Key on June 3, 2019

    \nCalculates the costs of constructing foundations for land-based wind projects *(items in brackets are not yet implemented)*:

    * Get number of turbines
    * Get duration of construction
    * Get daily hours of operation*  # todo: add to process diagram
    * Get season of construction*  # todo: add to process diagram
    * [Get region]
    * Get rotor diameter
    * Get hub height
    * Get turbine rating
    * Get buoyant foundation design flag
    * [Get seismic zone]
    * Get tower technology type
    * Get hourly weather data
    * [Get specific seasonal delays]
    * [Get long-term, site-specific climate data]
    * Get price data
    * Get labor rates
    * Get material prices for steel and concrete
    * [Use region to determine weather data]


    \n\nGiven below is the set of calculations carried out in this module:

    * Calculate the foundation loads using the rotor diameter, hub height, and turbine rating

    * Determine the foundation size based on the foundation loads, buoyant foundation design flag, and type of tower technology

    * Estimate the amount of material needed for foundation construction based on foundation size and number of turbines

    * Estimate the amount of time required to construct foundation based on foundation size, hours of operation, duration of construction, and number of turbines

    * Estimate the additional amount of time for weather delays (currently only assessing wind delays) based on hourly weather data, construction time, hours of operation, and season of construction

    * Estimate the amount of labor required for foundation construction based on foundation size, construction time, and weather delay
        * Calculate number of workers by crew type
        * Calculate man hours by crew type

    * Estimate the amount of equipment needed for foundation construction based on foundation size, construction time, and weather delay
        * Calculate number of equipment by equip type
        * Calculate equipment hours by equip type

    - Calculate the total foundation cost based on amount of equipment, amount of labor, amount of material, and price data.


    **Keys in the input dictionary are the following:**

    depth
        (int) depth of foundation [in m]


    component_data
        (pd.DataFrame) data frame with wind turbine component data

    def __init__(self, input_dict, output_dict, project_name):
        self.input_dict = input_dict
        self.output_dict = output_dict
        self.project_name = project_name


    num_turbines
        (int) total number of turbines in wind farm

    duration_construction
        (int) estimated construction time in months

    num_delays
        (int) Number of delay events

    avg_hours_per_delay
        (float) Average hours per delay event

    std_dev_hours_per_delay
        (float) Standard deviation from average hours per delay event

    delay_speed_m_per_s
        (float) wind speed above which weather delays kick in

    start_delay_hours
        (int)

    mission_time_hours
        (int)

    gust_wind_speed_m_per_s
        (float)

    wind_height_of_interest_m
        (int)

    wind_shear_exponent
        (float)

    season_construct
        list of seasons (like ['spring', 'summer']) for the construction.

    time_construct
        list of time windows for constructions. Use ['normal'] for a
        0800 to 1800 schedule 10 hour schedule. Use ['long'] for an
        overnight 1800 to 2359, 0000 to 0759 overnight schedule. Use
        ['normal', 'long'] for a 24-hour schedule.

    operational_hrs_per_day
        (float)


    material_price
        (pd.DataFrame) dataframe containing foundation cost related material prices

    rsmeans
        (pd.DataFrame) TODO: Formal definition for rsmeans?


    **Keys in the output dictionary are the following:**

    F_dead_kN_per_turbine
        (float) foundation dead load [in kN]

    F_horiz_kN_per_turbine
        (float) total lateral load [kN]

    M_tot_kN_m_per_turbine
        (float) Moment [kN.m]

    Radius_o_m
        (float) foundation radius based on overturning moment [in m]

    Radius_g_m
        (float) foundation radius based on gapping [in m]

    Radius_b_m
        (float) foundation radius based on bearing pressure [in m]

    Radius_m
        (float) largest foundation radius based on all three foundation design criteria: moment, gapping, bearing [in m]

    foundation_volume_concrete_m3_per_turbine
        (float) volume of a round, raft foundation [in m^3]

    steel_mass_short_ton
        (float) short tons of reinforcing steel

    material_needs_per_turbine
        (pd.DataFrame) table containing material needs info for -> Steel - rebar, Concrete 5000 psi, Excavated dirt, Backfill.

    operation_data
        (pd.DataFrame) TODO: What's the best one line definition for this?


    **TODO: Weather delay set of outputs -> ask Alicia for formal definitions of these keys.**

    total_foundation_cost
        (pd.DataFrame) summary of foundation costs (in USD) broken down into 4 main categories:
        1. Equipment Rental
        2. Labor
        3. Materials
        4. Mobilization
    """

    def __init__(self, input_dict, output_dict, project_name):
        """
        Parameters
        ----------
        input_dict : dict
            The input dictionary with key value pairs described in the
            class documentation

        output_dict : dict
            The output dictionary with key value pairs as found on the
            output documentation.
        """

        self.input_dict = input_dict
        self.output_dict = output_dict
        self.project_name = project_name

        # Constants used in FoundationCost class. Making this data private (hidden from outside of this class): #TODO: Change private variables to protected.
        self._kg_per_tonne = 1000
        self._cubicm_per_cubicft = 0.0283168
        self._steel_density = 9490  # kg / m^3
        self._cubicyd_per_cubicm = 1.30795
        self._ton_per_tonne = 0.907185

    def calculate_foundation_load(self, foundation_load_input_data, foundation_load_output_data):
        """

        Function to calculate foundation load.

        Parameters
        -------
        Int Section height m

        Surface area sq (in m^2)

        Coeff drag (installed)

        Lever arm m (in m)

        Multplier drag rotor

        Multiplier tower drag

        Mass tonne



        Returns
        -------
        Dead load [in N] -> F_dead_kN_per_turbine

        Lateral load [in N] -> F_horiz_kN_per_turbine

        Moment [N.m] -> M_tot_kN_m_per_turbine

        Foundation radius based on overturning moment [in m] -> Radius_o_m

        Foundation radius based on slipping [in m] -> Radius_s_m

        Foundation radius based on gapping [in m] -> Radius_g_m

        Foundation radius based on bearing pressure [in m] -> Radius_b_m

        Largest foundation radius based on all three foundation design criteria (moment, gapping, bearing [in m]) -> Radius_m

        Raises
        ------
        ValueError
            Raises a value error if r_bearing is calculated to be a negative value.
        """
        # set exposure constants
        a = 9.5
        z_g = 274.32

        # get section height
        z = foundation_load_input_data["Section height m"]

        # get cross-sectional area
        a_f = foundation_load_input_data["Surface area sq m"]

        # get coefficient of drag
        c_d = foundation_load_input_data["Coeff drag (installed)"]

        # get lever arm
        l = foundation_load_input_data["Lever arm m"]

        # get multipliers for tower and rotor
        multiplier_rotor = foundation_load_input_data["Multplier drag rotor"]
        multiplier_tower = foundation_load_input_data["Multiplier tower drag"]

        # calculate wind pressure
        k_z = 2.01 * (z / z_g) ** (2 / a)  # exposure factor
        k_d = 0.95  # wind directionality factor
        k_zt = 1  # topographic factor
        v = foundation_load_input_data["gust_velocity_m_per_s"]
        wind_pressure = 0.613 * k_z * k_zt * k_d * v ** 2

        # calculate wind loads on each tower component
        g = 0.85  # gust factor
        c_f = 0.6  # coefficient of force
        f_t = (wind_pressure * g * c_f * a_f) * multiplier_tower

        # calculate drag rotor
        rho = 1.225  # air density in kg/m^3
        f_r = (0.5 * rho * c_d * a_f * v ** 2) * multiplier_rotor

        f = f_t + f_r

        # calculate dead load in N
        g = 9.8  # m / s ^ 2
        f_dead = (
            sum(foundation_load_input_data["Mass tonne"]) * g * self._kg_per_tonne / 1.15
        )  # scaling factor to adjust dead load for uplift

        # calculate moment from each component at base of tower
        m_overturn = f * l

        # get total lateral load (N) and moment (N * m)
        f_lat = f.sum()  # todo: add f_lat (drag force) to output csv
        m_overturn = m_overturn.sum()

        # compare to moment from rated thrust
        rated_thrust = foundation_load_input_data["rated_thrust_N"]
        m_thrust = rated_thrust * max(l)
        m_tot = max(m_thrust, m_overturn)

        # compare lateral load to rated thrust
        f_horiz = max(f_lat, rated_thrust)

        # calculate foundation radius based on overturning moment
        vol_fraction_fill = 0.55
        vol_fraction_concrete = 1 - vol_fraction_fill
        safety_overturn = 1.5
        unit_weight_fill = 17.3e3  # in N / m^3
        unit_weight_concrete = 23.6e3  # in N / m^3
        bearing_pressure = foundation_load_input_data["bearing_pressure_n_m2"]
        p = [
            (
                np.pi
                * foundation_load_input_data["depth"]
                * (vol_fraction_fill * unit_weight_fill + vol_fraction_concrete * unit_weight_concrete)
            ),
            0,
            f_dead,
            -(safety_overturn * (m_tot + f_horiz * foundation_load_input_data["depth"])),
        ]
        r_overturn = np.roots(p)
        r_overturn = np.real(r_overturn[np.isreal(r_overturn)])[0]

        # calculate foundation radius based on slipping
        safety_slipping = 1.5
        friction_angle_soil = 25
        tangent_slip_angle = math.tan((friction_angle_soil * math.pi) / 180)
        slipping_force_with_sf = safety_slipping * f_lat
        # first check if slipping is already satisfied by dead weight
        if slipping_force_with_sf < (f_dead * tangent_slip_angle):
            r_slipping = 0
        else:
            # Calculate foundation radius based on slipping:
            r_slipping = (
                ((slipping_force_with_sf / tangent_slip_angle) - f_dead)
                / (
                    (vol_fraction_fill * unit_weight_fill + vol_fraction_concrete * unit_weight_concrete)
                    * math.pi
                    * foundation_load_input_data["depth"]
                )
            ) ** 0.5

        r_test_gapping = max(r_overturn, r_slipping)

        # calculate foundation radius based on gapping
        # check if gapping constrain is already satisfied - r / 3 < e
        foundation_vol = np.pi * r_test_gapping ** 2 * foundation_load_input_data["depth"]
        v_1 = (
            foundation_vol * (vol_fraction_fill * unit_weight_fill + vol_fraction_concrete * unit_weight_concrete)
            + f_dead
        )
        e = m_tot / v_1
        if (r_test_gapping / 3) < e:
            r_gapping = 0
        else:

            def r_g(x):
                foundation_vol = np.pi * x ** 2 * foundation_load_input_data["depth"]
                v_1 = (
                    foundation_vol
                    * (vol_fraction_fill * unit_weight_fill + vol_fraction_concrete * unit_weight_concrete)
                    + f_dead
                )
                e = m_tot / v_1
                return e * 3 - x

            result = root_scalar(r_g, method="brentq", bracket=[0.9 * r_overturn, 50], xtol=1e-4, maxiter=50)
            r_gapping = result.root
            if not result.converged:
                raise ValueError(
                    f"Warning {self.project_name} calculate_foundation_load r_gapping solve failed, {result.flag}"
                )

        r_test_bearing = max(r_test_gapping, r_gapping)

        # calculate foundation radius based on bearing pressure
        def r_b(x):
            foundation_vol = np.pi * r_test_bearing ** 2 * foundation_load_input_data["depth"]
            v_1 = (
                foundation_vol * (vol_fraction_fill * unit_weight_fill + vol_fraction_concrete * unit_weight_concrete)
                + f_dead
            )
            e = m_tot / v_1
            a_eff = v_1 / bearing_pressure
            return 2 * (x ** 2 - e * (x ** 2 - e ** 2) ** 0.5) - a_eff

        result = root_scalar(r_b, method="brentq", bracket=[0.9 * r_overturn, 50], xtol=1e-10, maxiter=50)
        r_bearing = result.root

        if not result.converged:
            raise ValueError(
                f"Warning {self.project_name} calculate_foundation_load r_bearing solve failed, {result.flag}"
            )

        # pick the largest foundation radius based on all 4 foundation design criteria: moment, gapping, bearing, slipping
        r_choosen = max(r_bearing, r_overturn, r_slipping, r_gapping)

        foundation_load_output_data["F_dead_kN_per_turbine"] = f_dead / 1e3
        foundation_load_output_data["F_horiz_kN_per_turbine"] = f_lat / 1e3
        foundation_load_output_data["M_tot_kN_m_per_turbine"] = m_tot / 1e3
        foundation_load_output_data["Radius_o_m"] = r_overturn
        foundation_load_output_data["Radius_s_m"] = r_slipping
        foundation_load_output_data["Radius_g_m"] = r_gapping
        foundation_load_output_data["Radius_b_m"] = r_bearing
        foundation_load_output_data["Radius_m"] = r_choosen

        return foundation_load_output_data

    def determine_foundation_size(self, foundation_size_input_data, foundation_size_output_data):
        """
        Function to calculate the volume of a round, raft foundation. Assumes foundation made of concrete with 1 m thickness.

        Parameters
        -------
        Largest foundation radius based on all three foundation design criteria: moment, gapping, bearing [in m] -> Radius_m [in m]

        depth of foundation [in m] -> depth


        Returns
        -------
        Foundation volume [in m^3] -> foundation_volume_concrete_m3_per_turbine

        """
        # TODO: still updating/fine-tuning foundation size equations for small DW (Parangat - Feb 27, 2020)
        r = float(foundation_size_output_data["Radius_m"])
        if foundation_size_input_data["turbine_rating_MW"] < 0.1:
            foundation_size_output_data["excavated_volume_m3"] = r * r * foundation_size_input_data["depth"]
            foundation_size_output_data["foundation_volume_concrete_m3_per_turbine"] = (
                foundation_size_output_data["excavated_volume_m3"] * 0.45
            )
        else:
            foundation_size_output_data["excavated_volume_m3"] = (
                np.pi * (r + 0.5) ** 2 * foundation_size_input_data["depth"]
            )

            # only compute the portion of the foundation that is composed of concrete (45% concrete; other portion is
            # backfill); TODO: Add to sphinx -> (volume excavated = pi*(r_pick + .5m)^2 this assumes vertical sides which
            #  does not reflect reality as OSHA requires benched sides over 3’)
            foundation_size_output_data["foundation_volume_concrete_m3_per_turbine"] = (
                np.pi * r ** 2 * foundation_size_input_data["depth"] * 0.45
            )

        return foundation_size_output_data

    def estimate_material_needs_per_turbine(
        self, material_needs_per_turbine_input_data, material_needs_per_turbine_output_data
    ):
        """
        Function to estimate amount of material based on foundation size and number of turbines.


        Parameters
        -------
        Foundation concrete volume [in m^3] -> foundation_volume_concrete_m3_per_turbine


        Returns
        -------

        (Returns pd.DataFrame) material_needs_per_turbine


        """

        steel_mass_short_ton_per_turbine = (
            material_needs_per_turbine_output_data["foundation_volume_concrete_m3_per_turbine"]
            * 0.012
            * self._steel_density
            / self._kg_per_tonne
        )
        concrete_volume_cubic_yards_per_turbine = (
            material_needs_per_turbine_output_data["foundation_volume_concrete_m3_per_turbine"]
            * 0.985
            * self._cubicyd_per_cubicm
        )

        # Assign values to output dictionary:
        material_needs_per_turbine_output_data["material_needs_per_turbine"] = pd.DataFrame(
            [
                ["Steel - rebar", steel_mass_short_ton_per_turbine, "ton (short)"],
                ["Concrete 5000 psi", concrete_volume_cubic_yards_per_turbine, "cubic yards"],
                [
                    "Excavated dirt",
                    material_needs_per_turbine_output_data["excavated_volume_m3"] * self._cubicyd_per_cubicm,
                    "cubic_yards",
                ],
                [
                    "Backfill",
                    material_needs_per_turbine_output_data["excavated_volume_m3"] * self._cubicyd_per_cubicm,
                    "cubic_yards",
                ],
            ],
            columns=["Material type ID", "Quantity of material", "Units"],
        )

        material_needs_per_turbine_output_data["steel_mass_short_ton_per_turbine"] = steel_mass_short_ton_per_turbine
        return material_needs_per_turbine_output_data["material_needs_per_turbine"]

    def estimate_construction_time(self, construction_time_input_data, construction_time_output_data):
        """
        Function to estimate construction time on per turbine basis. TODO: What's a better definition of this function. It's task is to return a pd.DataFrame (operation_data).

        Parameters
        -------
        duration_construction

        pd.DataFrame
            rsmeans

        pd.DataFrame
            material_needs_per_turbine



        Returns
        -------

        (pd.DataFrame) operation_data

        """

        foundation_construction_time = construction_time_input_data["construct_duration"] * 1 / 3
        # throughput_operations = construction_time_input_data['throughput_operations']
        throughput_operations = construction_time_input_data["rsmeans"]
        material_needs_per_turbine = construction_time_output_data["material_needs_per_turbine"]
        quantity_materials_entire_farm = (
            material_needs_per_turbine["Quantity of material"] * construction_time_input_data["num_turbines"]
        )

        # Calculations for estimate construction time will be on entire wind farm basis:
        construction_time_output_data["material_needs_entire_farm"] = material_needs_per_turbine.copy()
        material_needs_entire_farm = construction_time_output_data["material_needs_entire_farm"]
        material_needs_entire_farm["Quantity of material"] = quantity_materials_entire_farm
        if construction_time_input_data["turbine_rating_MW"] <= 0.1:
            operation_data = throughput_operations.where(
                throughput_operations["Module"] == "Small DW Foundations"
            ).dropna(thresh=4)
        else:
            operation_data = throughput_operations.where(throughput_operations["Module"] == "Foundations").dropna(
                thresh=4
            )

        # operation data for entire wind farm:
        operation_data = pd.merge(material_needs_entire_farm, operation_data, on=["Material type ID"], how="outer")
        operation_data["Number of days"] = operation_data["Quantity of material"] / operation_data["Daily output"]
        operation_data["Number of crews"] = np.ceil(
            (operation_data["Number of days"] / 30) / foundation_construction_time
        )

        alpha = operation_data[operation_data["Type of cost"] == "Labor"]
        operation_data_id_days_crews_workers = alpha[
            ["Operation ID", "Number of days", "Number of crews", "Number of workers"]
        ]

        # if more than one crew needed to complete within construction duration then assume that all construction happens
        # within that window and use that timeframe for weather delays; if not, use the number of days calculated
        operation_data["time_construct_bool"] = operation_data["Number of days"] > foundation_construction_time * 30
        boolean_dictionary = {True: foundation_construction_time * 30, False: np.NAN}
        operation_data["time_construct_bool"] = operation_data["time_construct_bool"].map(boolean_dictionary)
        operation_data["Time construct days"] = operation_data[["time_construct_bool", "Number of days"]].min(axis=1)
        num_days = operation_data["Time construct days"].max()

        construction_time_output_data["operation_data_id_days_crews_workers"] = operation_data_id_days_crews_workers
        construction_time_output_data["operation_data_entire_farm"] = operation_data

        # pull out management data #TODO: Add this cost to Labor cost next
        if construction_time_input_data["turbine_rating_MW"] > 0.1:
            crew_cost = self.input_dict["crew_cost"]
            crew = self.input_dict["crew"][self.input_dict["crew"]["Crew type ID"].str.contains("M0")]
            management_crew = pd.merge(crew_cost, crew, on=["Labor type ID"])
            management_crew = management_crew.assign(
                per_diem_total=management_crew["Per diem USD per day"] * management_crew["Number of workers"] * num_days
            )
            management_crew = management_crew.assign(
                hourly_costs_total=management_crew["Hourly rate USD per hour"]
                * self.input_dict["hour_day"][self.input_dict["time_construct"]]
                * num_days
            )
            management_crew = management_crew.assign(
                total_crew_cost_before_wind_delay=management_crew["per_diem_total"]
                + management_crew["hourly_costs_total"]
            )
            self.output_dict["management_crew"] = management_crew
            self.output_dict["managament_crew_cost_before_wind_delay"] = management_crew[
                "total_crew_cost_before_wind_delay"
            ].sum()
        else:
            self.output_dict["managament_crew_cost_before_wind_delay"] = 0

        return construction_time_output_data["operation_data_entire_farm"]

    def calculate_weather_delay(self, weather_delay_input_data, weather_delay_output_data):
        """
        Function to calculate wind delay for foundations.

        Keys in weather_delay_input_data
        --------------------------------
        weather_window

        duration_construction

        start_delay

        critical_wind_speed

        operational_hrs_per_day

        height_interest

        wind_shear_exponent
        """

        # construct WeatherDelay module
        WD(weather_delay_input_data, weather_delay_output_data)

        # compute weather delay
        wind_delay = pd.DataFrame(weather_delay_output_data["wind_delays"])

        # if greater than 4 hour delay, then shut down for full day (10 hours)
        wind_delay[(wind_delay > 4)] = 10
        weather_delay_output_data["wind_delay_time"] = float(wind_delay.sum())

        return weather_delay_output_data

    def calculate_costs(self, calculate_costs_input_dict, calculate_costs_output_dict):
        """
        Function to calculate the total foundation cost.

        Keys in input dictionary
        ------------------------
        pd.DataFrame
            material_needs_per_turbine

        pd.DataFrame
            material_price

        pd.DataFrame
            operation_data

        wind_delay_time

        operational_hrs_per_day

        wind_multiplier

        pd.DataFrame
            rsmeans


        Returns
        -------

        (pd.DataFrame) total_foundation_cost


        """

        material_vol_entire_farm = calculate_costs_output_dict["material_needs_entire_farm"]
        material_price = calculate_costs_input_dict["material_price"]

        material_data_entire_farm = pd.merge(material_vol_entire_farm, material_price, on=["Material type ID"])
        material_data_entire_farm["Cost USD"] = material_data_entire_farm["Quantity of material"] * pd.to_numeric(
            material_data_entire_farm["Material price USD per unit"]
        )  # material data on a total wind farm basis

        operation_data = calculate_costs_output_dict["operation_data_entire_farm"]

        wind_delay = calculate_costs_output_dict["wind_delay_time"]

        wind_delay_fraction = (wind_delay / calculate_costs_input_dict["operational_hrs_per_day"]) / operation_data[
            "Time construct days"
        ].max(skipna=True)
        # check if wind_delay_fraction is greater than 1, which would mean weather delays are longer than they can possibily be for the input data
        if wind_delay_fraction > 1:
            raise ValueError("{}: Error: Wind delay greater than 100%".format(type(self).__name__))
        wind_multiplier = 1 / (1 - wind_delay_fraction)
        calculate_costs_output_dict["wind_multiplier"] = wind_multiplier

        rsmeans = calculate_costs_input_dict["rsmeans"]
        if calculate_costs_input_dict["turbine_rating_MW"] > 0.1:
            rsmeans = rsmeans.where(rsmeans["Module"] == "Foundations").dropna(thresh=4)
        else:
            rsmeans = rsmeans.where(rsmeans["Module"] == "Small DW Foundations").dropna(thresh=4)

        labor_equip_data = pd.merge(material_vol_entire_farm, rsmeans, on=["Material type ID"])

        # Create foundation cost dataframe
        foundation_cost = pd.DataFrame(columns=["Type of cost", "Cost USD", "Phase of construction"])

        # Calculate per diem
        per_diem = (
            operation_data["Number of workers"]
            * operation_data["Number of crews"]
            * (operation_data["Time construct days"] + np.ceil(operation_data["Time construct days"] / 7))
            * calculate_costs_input_dict["rsmeans_per_diem"]
        )
        where_are_na_ns = np.isnan(per_diem)
        per_diem[where_are_na_ns] = 0
        labor_equip_data["Cost USD"] = (
            labor_equip_data["Quantity of material"]
            * labor_equip_data["Rate USD per unit"]
            * calculate_costs_input_dict["overtime_multiplier"]
            + per_diem
            + calculate_costs_output_dict["managament_crew_cost_before_wind_delay"]
        ) * wind_multiplier
        self.output_dict["labor_equip_data"] = labor_equip_data

        # EQUIPMENT COST
        # Create equipment costs row to be appended to foundation_cost
        equipment_dataframe = labor_equip_data[labor_equip_data["Type of cost"].str.match("Equipment rental")]

        equipment_cost_usd_without_delay = (
            equipment_dataframe["Quantity of material"]
            * equipment_dataframe["Rate USD per unit"]
            * calculate_costs_input_dict["overtime_multiplier"]
            + per_diem
        )
        equipment_cost_usd_with_weather_delays = equipment_cost_usd_without_delay.sum() * wind_multiplier
        equipment_costs = pd.DataFrame(
            [["Equipment rental", equipment_cost_usd_with_weather_delays, "Foundation"]],
            columns=["Type of cost", "Cost USD", "Phase of construction"],
        )

        # LABOR COST
        # Create labor costs row to be appended to foundation_cost
        labor_dataframe = labor_equip_data[labor_equip_data["Type of cost"].str.match("Labor")]
        labor_cost_usd_without_management = (
            labor_dataframe["Quantity of material"]
            * labor_dataframe["Rate USD per unit"]
            * calculate_costs_input_dict["overtime_multiplier"]
            + per_diem
        )
        labor_cost_usd_with_management = (
            labor_cost_usd_without_management.sum()
            + calculate_costs_output_dict["managament_crew_cost_before_wind_delay"]
        )
        labor_cost_usd_with_management_plus_weather_delays = labor_cost_usd_with_management * wind_multiplier
        labor_costs = pd.DataFrame(
            [["Labor", labor_cost_usd_with_management_plus_weather_delays, "Foundation"]],
            columns=["Type of cost", "Cost USD", "Phase of construction"],
        )

        # MATERIAL COST
        material_cost_dataframe = pd.DataFrame(columns=["Operation ID", "Type of cost", "Cost USD"])
        material_cost_dataframe["Operation ID"] = material_data_entire_farm["Material type ID"]
        material_cost_dataframe["Type of cost"] = "Materials"
        material_cost_dataframe["Cost USD"] = material_data_entire_farm["Cost USD"]
        material_costs_sum = material_cost_dataframe["Cost USD"].sum()
        material_costs = pd.DataFrame(
            [["Materials", material_costs_sum, "Foundation"]],
            columns=["Type of cost", "Cost USD", "Phase of construction"],
        )

        # Append all cost items to foundation_cost
        foundation_cost = foundation_cost.append(equipment_costs)
        foundation_cost = foundation_cost.append(labor_costs)
        foundation_cost = foundation_cost.append(material_costs)

        # Calculate mobilization cost as percentage of total foundation cost and add to foundation_cost
        # Assumed 5% of total foundation cost and add to foundation_cost for utility scale plant
        # A function of turbine size for distributed wind (< 10 turbines)
        if calculate_costs_input_dict["num_turbines"] > 10:
            mobilization_cost = foundation_cost["Cost USD"].sum() * 0.05
        else:
            if calculate_costs_input_dict["turbine_rating_MW"] < 0.1:
                # Zero since mobilization cost of equipment is included in the equipment rental cost
                mobilization_cost = 0
            else:
                # There is mobilization cost for 0-10 turbines 100+ kW in rating.
                num_turbines = calculate_costs_input_dict["num_turbines"]
                rating = calculate_costs_input_dict["turbine_rating_MW"]
                mobilization_multipler = self.mobilization_cost_multiplier(rating)
                mobilization_cost = foundation_cost["Cost USD"].sum() / num_turbines * mobilization_multipler

        mob_cost = pd.DataFrame(
            [["Mobilization", mobilization_cost, "Foundation"]],
            columns=["Type of cost", "Cost USD", "Phase of construction"],
        )

        foundation_cost = foundation_cost.append(mob_cost)

        # todo: we add a separate tab in the output file for costs (all costs will be the same format but it's a different format than other data)
        # columns in cost tab would include project_id, module, operation_id, type_of_cost, total_or_per_turbine, cost_usd
        # an example row for this type of output would be "project1, FoundationCost, Rebar installation, Labor, total, 2.371127e+06"
        # total_foundation_cost = foundation_cost.groupby(by=['Type of cost']).sum().reset_index()
        # total_foundation_cost['Phase of construction'] = 'Foundations'
        # total_cost_summed_foundation = total_foundation_cost.sum(numeric_only=True)[0] # todo: add total_cost_summed_foundation to output dict

        total_foundation_cost = foundation_cost
        calculate_costs_output_dict["total_foundation_cost"] = total_foundation_cost

        self.output_dict["labor_equip_data"] = labor_equip_data

        return total_foundation_cost

    def outputs_for_detailed_tab(self, input_dict, output_dict):
        """
        Creates a list of dictionaries which can be used on their own or
        used to make a dataframe.

        Must be called after self.run_module()

        Returns
        -------
        list(dict)
            A list of dicts, with each dict representing a row of the data.
        """

        # Note that some values are cast with float() so that XlsxWriter
        # (the library depended on by XlsxGenerator) can output them as
        # numbers. XlsxWriter, interestingly, cannot handle np.float32()
        # types.

        result = []
        module = type(self).__name__
        result.append(
            {
                "unit": "",
                "type": "variable",
                "variable_df_key_col_name": "wind_multiplier",
                "value": float(self.output_dict["wind_multiplier"]),
            }
        )
        result.append(
            {
                "unit": "kN",
                "type": "variable",
                "variable_df_key_col_name": "F_dead",
                "value": float(self.output_dict["F_dead_kN_per_turbine"]),
            }
        )
        result.append(
            {
                "unit": "kN",
                "type": "variable",
                "variable_df_key_col_name": "F_horiz",
                "value": float(self.output_dict["F_horiz_kN_per_turbine"]),
            }
        )
        result.append(
            {
                "unit": "kN_m",
                "type": "variable",
                "variable_df_key_col_name": "M_tot_kN",
                "value": float(self.output_dict["M_tot_kN_m_per_turbine"]),
            }
        )
        result.append(
            {
                "unit": "m",
                "type": "variable",
                "variable_df_key_col_name": "Radius_o",
                "value": float(self.output_dict["Radius_o_m"]),
            }
        )
        result.append(
            {
                "unit": "m",
                "type": "variable",
                "variable_df_key_col_name": "Radius_g",
                "value": float(self.output_dict["Radius_g_m"]),
            }
        )
        result.append(
            {
                "unit": "m",
                "type": "variable",
                "variable_df_key_col_name": "Radius_b",
                "value": float(self.output_dict["Radius_b_m"]),
            }
        )
        result.append(
            {
                "unit": "m",
                "type": "variable",
                "variable_df_key_col_name": "Radius",
                "value": float(self.output_dict["Radius_m"]),
            }
        )
        result.append(
            {
                "unit": "short_ton",
                "type": "variable",
                "variable_df_key_col_name": "steel_mass_short_ton_per_turbine",
                "value": self.output_dict["steel_mass_short_ton_per_turbine"],
            }
        )
        # foundation_volume_concrete_m3_per_turbine
        result.append(
            {
                "unit": "m^3",
                "type": "variable",
                "variable_df_key_col_name": "foundation_volume_concrete_m3_per_turbine",
                "value": self.output_dict["foundation_volume_concrete_m3_per_turbine"],
            }
        )

        for row in self.output_dict["operation_data_id_days_crews_workers"].itertuples():
            dashed_row = "{}-{}-{}-{}".format(row[1], math.ceil(row[2]), row[3], row[4])
            result.append(
                {
                    "unit": "",
                    "type": "dataframe",
                    "variable_df_key_col_name": "operation_data: Operation ID-Number of days-Number of crews-Number of workers",
                    "value": dashed_row,
                }
            )

        for row in self.output_dict["material_needs_per_turbine"].itertuples():
            # This must be formatted in Python
            dashed_row = "{}-{}-{:.2e}".format(row[0], row[1], row[2])
            result.append(
                {
                    "unit": row[3],
                    "type": "dataframe",
                    "variable_df_key_col_name": "material_needs_per_turbine: {}".format(
                        "-".join(self.output_dict["material_needs_per_turbine"].columns[:-1])
                    ),
                    "value": dashed_row,
                }
            )

        for row in self.output_dict["total_foundation_cost"].itertuples():
            dashed_row = "{} <--> {} <--> {}".format(row[1], row[3], math.ceil(row[2]))
            result.append(
                {
                    "unit": "",
                    "type": "dataframe",
                    "variable_df_key_col_name": "Type of Cost <--> Phase of Construction <--> Cost in USD ",
                    "value": dashed_row,
                    "last_number": row[2],
                }
            )

        for _dict in result:
            _dict["project_id_with_serial"] = self.project_name
            _dict["module"] = module

        self.output_dict["foundation_cost_csv"] = result
        return result

    def outputs_for_module_type_operation(self, input_dict, output_dict):
        result = []
        module = type(self).__name__

        costs_by_module_type_operation = self.output_dict["labor_equip_data"]
        for _, row in costs_by_module_type_operation.iterrows():
            _dict = dict()
            row = row.to_dict()
            _dict["operation_id"] = row["Operation ID"]
            _dict["type_of_cost"] = row["Type of cost"]
            _dict["cost"] = row["Cost USD"]
            result.append(_dict)

        for _dict in result:
            _dict["project_id_with_serial"] = self.project_name
            _dict["module"] = module
            _dict["total_or_turbine"] = "total"

        self.output_dict["foundation_module_type_operation"] = result
        return result

    def run_module(self):
        """
        Runs the FoundationCost module and populates the IO dictionaries with calculated values.

        Parameters
        ----------
        <None>

        Returns
        -------
        tuple
            First element of tuple contains a 0 or 1. 0 means no errors happened and
            1 means an error happened and the module failed to run. The second element
            either returns a 0 if the module ran successfully, or it returns the error
            raised that caused the failure.

        """
        try:
            self.calculate_foundation_load(self.input_dict, self.output_dict)  # Returns foundation load
            self.determine_foundation_size(self.input_dict, self.output_dict)  # Returns foundation volume
            self.estimate_material_needs_per_turbine(self.input_dict, self.output_dict)  # Returns material volume
            operation_data = self.estimate_construction_time(
                self.input_dict, self.output_dict
            )  # Estimates construction time

            # pull only global inputs for weather delay from input_dict
            weather_data_keys = ("wind_shear_exponent", "weather_window")

            # specify foundation-specific weather delay inputs
            self.weather_input_dict = dict(
                [(i, self.input_dict[i]) for i in self.input_dict if i in set(weather_data_keys)]
            )
            self.weather_input_dict[
                "start_delay_hours"
            ] = 0  # assume zero start for when foundation construction begins (start at beginning of construction time)
            self.weather_input_dict["critical_wind_speed_m_per_s"] = self.input_dict[
                "critical_speed_non_erection_wind_delays_m_per_s"
            ]
            self.weather_input_dict["wind_height_of_interest_m"] = self.input_dict[
                "critical_height_non_erection_wind_delays_m"
            ]

            # duration_construction is in units of days
            # duration_construction_months is in units of months
            days_per_month = 30
            duration_construction = operation_data["Time construct days"].max(skipna=True)
            duration_construction_months = duration_construction / days_per_month
            self.output_dict["foundation_construction_months"] = duration_construction_months

            # compute and specify weather delay mission time for roads
            operational_hrs_per_day = self.input_dict["hour_day"][self.input_dict["time_construct"]]
            mission_time_hrs = duration_construction * operational_hrs_per_day
            self.weather_input_dict["mission_time_hours"] = mission_time_hrs
            self.input_dict["operational_hrs_per_day"] = operational_hrs_per_day

            self.calculate_weather_delay(self.weather_input_dict, self.output_dict)
            self.calculate_costs(self.input_dict, self.output_dict)
            self.outputs_for_detailed_tab(self.input_dict, self.output_dict)
            # self.output_dict['labor_equip_data']
            # self.output_dict['foundation_module_type_operation'] = self.outputs_for_module_type_operation(self.input_dict, self.output_dict)
            self.output_dict["foundation_module_type_operation"] = self.outputs_for_costs_by_module_type_operation(
                input_df=self.output_dict["total_foundation_cost"], project_id=self.project_name, total_or_turbine=True
            )
            return 0, 0  # module ran successfully
        except Exception as error:
            traceback.print_exc()
            print(f"Fail {self.project_name} FoundationCost")
            return 1, error  # module did not run successfully



"""
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import math
import unittest
from os import path

import numpy as np
from wisdem.ccblade.ccblade import CCBlade, CCAirfoil


class TestNREL5MW(unittest.TestCase):
    def setUp(self):

        # geometry
        Rhub = 1.5
        Rtip = 63.0

        r = np.array(
            [
                2.8667,
                5.6000,
                8.3333,
                11.7500,
                15.8500,
                19.9500,
                24.0500,
                28.1500,
                32.2500,
                36.3500,
                40.4500,
                44.5500,
                48.6500,
                52.7500,
                56.1667,
                58.9000,
                61.6333,
            ]
        )
        chord = np.array(
            [
                3.542,
                3.854,
                4.167,
                4.557,
                4.652,
                4.458,
                4.249,
                4.007,
                3.748,
                3.502,
                3.256,
                3.010,
                2.764,
                2.518,
                2.313,
                2.086,
                1.419,
            ]
        )
        theta = np.array(
            [
                13.308,
                13.308,
                13.308,
                13.308,
                11.480,
                10.162,
                9.011,
                7.795,
                6.544,
                5.361,
                4.188,
                3.125,
                2.319,
                1.526,
                0.863,
                0.370,
                0.106,
            ]
        )
        B = 3  # number of blades

        # atmosphere
        rho = 1.225
        mu = 1.81206e-5

        afinit = CCAirfoil.initFromAerodynFile  # just for shorthand
        basepath = path.join(path.dirname(path.realpath(__file__)), "../../../examples/_airfoil_files")

        # load all airfoils
        airfoil_types = [0] * 8
        airfoil_types[0] = afinit(path.join(basepath, "Cylinder1.dat"))
        airfoil_types[1] = afinit(path.join(basepath, "Cylinder2.dat"))
        airfoil_types[2] = afinit(path.join(basepath, "DU40_A17.dat"))
        airfoil_types[3] = afinit(path.join(basepath, "DU35_A17.dat"))
        airfoil_types[4] = afinit(path.join(basepath, "DU30_A17.dat"))
        airfoil_types[5] = afinit(path.join(basepath, "DU25_A17.dat"))
        airfoil_types[6] = afinit(path.join(basepath, "DU21_A17.dat"))
        airfoil_types[7] = afinit(path.join(basepath, "NACA64_A17.dat"))

        # place at appropriate radial stations
        af_idx = [0, 0, 1, 2, 3, 3, 4, 5, 5, 6, 6, 7, 7, 7, 7, 7, 7]

        af = [0] * len(r)
        for i in range(len(r)):
            af[i] = airfoil_types[af_idx[i]]

        tilt = -5.0
        precone = 2.5
        yaw = 0.0

        # create CCBlade object
        self.rotor = CCBlade(r, chord, theta, af, Rhub, Rtip, B, rho, mu, precone, tilt, yaw, shearExp=0.2, hubHt=90.0)

    def test_thrust_torque(self):

        Uinf = np.array([3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25])
        Omega = np.array(
            [
                6.972,
                7.183,
                7.506,
                7.942,
                8.469,
                9.156,
                10.296,
                11.431,
                11.890,
                12.100,
                12.100,
                12.100,
                12.100,
                12.100,
                12.100,
                12.100,
                12.100,
                12.100,
                12.100,
                12.100,
                12.100,
                12.100,
                12.100,
            ]
        )
        pitch = np.array(
            [
                0.000,
                0.000,
                0.000,
                0.000,
                0.000,
                0.000,
                0.000,
                0.000,
                0.000,
                3.823,
                6.602,
                8.668,
                10.450,
                12.055,
                13.536,
                14.920,
                16.226,
                17.473,
                18.699,
                19.941,
                21.177,
                22.347,
                23.469,
            ]
        )

        Pref = np.array(
            [
                42.9,
                188.2,
                427.9,
                781.3,
                1257.6,
                1876.2,
                2668.0,
                3653.0,
                4833.2,
                5296.6,
                5296.6,
                5296.6,
                5296.6,
                5296.6,
                5296.6,
                5296.6,
                5296.6,
                5296.7,
                5296.6,
                5296.7,
                5296.6,
                5296.6,
                5296.7,
            ]
        )
        Tref = np.array(
            [
                171.7,
                215.9,
                268.9,
                330.3,
                398.6,
                478.0,
                579.2,
                691.5,
                790.6,
                690.0,
                608.4,
                557.9,
                520.5,
                491.2,
                467.7,
                448.4,
                432.3,
                418.8,
                406.7,
                395.3,
                385.1,
                376.7,
                369.3,
            ]
        )
        Qref = np.array(
            [
                58.8,
                250.2,
                544.3,
                939.5,
                1418.1,
                1956.9,
                2474.5,
                3051.1,
                3881.3,
                4180.1,
                4180.1,
                4180.1,
                4180.1,
                4180.1,
                4180.1,
                4180.1,
                4180.1,
                4180.1,
                4180.1,
                4180.1,
                4180.1,
                4180.1,
                4180.1,
            ]
        )

        m_rotor = 110.0  # kg
        g = 9.81
        tilt = 5 * math.pi / 180.0
        Tref -= m_rotor * g * math.sin(tilt)  # remove weight of rotor that is included in reported results

        outputs, derivs = self.rotor.evaluate(Uinf, Omega, pitch)
        P, T, Q = [outputs[key] for key in ("P", "T", "Q")]

        # import matplotlib.pyplot as plt
        # plt.plot(Uinf, P/1e6)
        # plt.plot(Uinf, Pref/1e3)
        # plt.figure()
        # plt.plot(Uinf, T/1e6)
        # plt.plot(Uinf, Tref/1e3)
        # plt.show()

        idx = Uinf < 15
        np.testing.assert_allclose(Q[idx] / 1e6, Qref[idx] / 1e3, atol=0.15)
        np.testing.assert_allclose(P[idx] / 1e6, Pref[idx] / 1e3, atol=0.2)  # within 0.2 of 1MW
        np.testing.assert_allclose(T[idx] / 1e6, Tref[idx] / 1e3, atol=0.15)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestNREL5MW))
    return suite


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



import time

import numpy as np


# base class for MoorPy exceptions
class Error(Exception):
    """ Base class for MoorPy exceptions"""

    pass


# Catenary error class
class CatenaryError(Error):
    """Derived error class for catenary function errors. Contains an error message."""

    def __init__(self, message):
        self.message = message


# Line Object error class
class LineError(Error):
    """Derived error class for Line object errors. Contains an error message and the line number with the error."""

    def __init__(self, num, message):
        self.line_num = num
        self.message = message


# Solve error class for any solver process
class SolveError(Error):
    """Derived error class for various solver errors. Contains an error message"""

    def __init__(self, message):
        self.message = message


# Generic MoorPy error
class MoorPyError(Error):
    """Derived error class for MoorPy. Contains an error message"""

    def __init__(self, message):
        self.message = str(message)


def printMat(mat):
    """Prints a matrix to a format that is specified

    Parameters
    ----------
    mat : array
        Any matrix that is to be printed.

    Returns
    -------
    None.

    """
    for i in range(mat.shape[0]):
        print("\t".join(["{:+8.3e}"] * mat.shape[1]).format(*mat[i, :]))


def printVec(vec):
    """Prints a vector to a format that is specified

    Parameters
    ----------
    vec : array
        Any vector that is to be printed.

    Returns
    -------
    None.

    """
    print("\t".join(["{:+9.4e}"] * len(vec)).format(*vec))


def getH(r):
    """function gets the alternator matrix, H, that when multiplied with a vector,
    returns the cross product of r and that vector

    Parameters
    ----------
    r : array
        the position vector that another vector is from a point of interest.

    Returns
    -------
    H : matrix
        the alternator matrix for the size-3 vector, r.

    """

    H = np.array([[0, r[2], -r[1]], [-r[2], 0, r[0]], [r[1], -r[0], 0]])
    return H


def rotationMatrix(x3, x2, x1):
    """Calculates a rotation matrix based on order-z,y,x instrinsic (tait-bryan?) angles, meaning
    they are about the ROTATED axes. (rotation about z-axis would be (0,0,theta) )

    Parameters
    ----------
    x3, x2, x1: floats
        The angles that the rotated axes are from the nonrotated axes. Normally roll,pitch,yaw respectively. [rad]

    Returns
    -------
    R : matrix
        The rotation matrix
    """
    # initialize the sines and cosines
    s1 = np.sin(x1)
    c1 = np.cos(x1)
    s2 = np.sin(x2)
    c2 = np.cos(x2)
    s3 = np.sin(x3)
    c3 = np.cos(x3)

    # create the rotation matrix
    R = np.array(
        [
            [c1 * c2, c1 * s2 * s3 - c3 * s1, s1 * s3 + c1 * c3 * s2],
            [c2 * s1, c1 * c3 + s1 * s2 * s3, c3 * s1 * s2 - c1 * s3],
            [-s2, c2 * s3, c2 * c3],
        ]
    )

    return R


def rotatePosition(rRelPoint, rot3):
    """Calculates the new position of a point by applying a rotation (rotates a vector by three angles)

    Parameters
    ----------
    rRelPoint : array
        x,y,z coordinates of a point relative to a local frame [m]
    rot3 : array
        Three angles that describe the difference between the local frame and the global frame/ Normally roll,pitch,yaw. [rad]

    Returns
    -------
    rRel : array
        The relative rotated position of the point about the local frame [m]
    """

    # get rotation matrix from three provided angles
    RotMat = rotationMatrix(rot3[0], rot3[1], rot3[2])

    # find location of point in unrotated reference frame about reference point
    rRel = np.matmul(RotMat, rRelPoint)

    return rRel


def transformPosition(rRelPoint, r6):
    """Calculates the position of a point based on its position relative to translated and rotated 6DOF body

    Parameters
    ----------
    rRelPoint : array
        x,y,z coordinates of a point relative to a local frame [m]
    r6 : array
        6DOF position vector of the origin of the local frame, in the global frame coorindates [m]

    Returns
    -------
    rAbs : array
        The absolute position of the point about the global frame [m]
    """
    # note: r6 should be in global orientation frame

    # absolute location = rotation of relative position + absolute position of reference point
    rAbs = rotatePosition(rRelPoint, r6[3:]) + r6[:3]

    return rAbs


def translateForce3to6DOF(r, Fin):
    """Takes in a position vector and a force vector (applied at the positon), and calculates
    the resulting 6-DOF force and moment vector.

    Parameters
    ----------
    r : array
        x,y,z coordinates at which force is acting [m]
    Fin : array
        x,y,z components of force [N]

    Returns
    -------
    Fout : array
        The resulting force and moment vector [N, Nm]
    """

    # initialize output vector as same dtype as input vector (to support both real and complex inputs)
    Fout = np.zeros(6, dtype=Fin.dtype)

    # set the first three elements of the output vector the same as the input vector
    Fout[:3] = Fin

    # set the last three elements of the output vector as the cross product of r and Fin
    Fout[3:] = np.cross(r, Fin)

    return Fout


def set_axes_equal(ax):
    """Sets 3D plot axes to equal scale

    Parameters
    ----------
    ax : matplotlib.pyplot axes
        the axes that are to be set equal in scale to each other.

    Returns
    -------
    None.

    """

    rangex = np.diff(ax.get_xlim3d())[0]
    rangey = np.diff(ax.get_ylim3d())[0]
    rangez = np.diff(ax.get_zlim3d())[0]

    ax.set_box_aspect([rangex, rangey, rangez])  # note: this may require a matplotlib update


def dsolve2(
    eval_func,
    X0,
    Ytarget=[],
    step_func=None,
    args=[],
    tol=0.0001,
    maxIter=20,
    Xmin=[],
    Xmax=[],
    a_max=2.0,
    dX_last=[],
    stepfac=4,
    display=0,
):
    """
    PARAMETERS
    ----------
    eval_func : function
        function to solve (will be passed array X, and must return array Y of same size)
    X0 : array
        initial guess of X
    Ytarget : array (optional)
        target function results (Y), assumed zero if not provided
    stp_func : function (optional)
        function use for adjusting the variables (computing dX) each step.
        If not provided, Netwon's method with finite differencing is used.
    args : list
        A list of variables (e.g. the system object) to be passed to both the eval_func and step_func
    tol : float or array
        If scalar, the*relative* convergence tolerance (applied to step size components, dX).
        If an array, must be same size as X, and specifies an absolute convergence threshold for each variable.
    Xmin, Xmax
        Bounds. by default start bounds at infinity
    a_max
        maximum step size acceleration allowed
    dX_last
        Used if you want to dictate the initial step size/direction based on a previous attempt
    """
    success = False
    start_time = time.time()
    # process inputs and format as arrays in case they aren't already

    X = np.array(X0, dtype=np.float_)  # start off design variable
    N = len(X)

    Xs = np.zeros([maxIter, N])  # make arrays to store X and error results of the solve
    Es = np.zeros([maxIter, N])
    dXlist = np.zeros([maxIter, N])
    dXlist2 = np.zeros([maxIter, N])

    # check the target Y value input
    if len(Ytarget) == N:
        Ytarget = np.array(Ytarget, dtype=np.float_)
    elif len(Ytarget) == 0:
        Ytarget = np.zeros(N, dtype=np.float_)
    else:
        raise TypeError("Ytarget must be of same length as X0")

    # ensure all tolerances are positive
    if np.isscalar(tol) and tol <= 0.0:
        raise ValueError("tol value passed to dsovle2 must be positive")
    elif not np.isscalar(tol) and any(tol <= 0):
        raise ValueError("every tol entry passed to dsovle2 must be positive")

    # if a step function wasn't provided, provide a default one
    if step_func == None:
        if display > 1:
            print("Using default finite difference step func")

        def step_func(X, args, Y, oths, Ytarget, err, tols, iter, maxIter):
            """ this now assumes tols passed in is a vector and are absolute quantities"""
            J = np.zeros([N, N])  # Initialize the Jacobian matrix that has to be a square matrix with nRows = len(X)

            for i in range(
                N
            ):  # Newton's method: perturb each element of the X variable by a little, calculate the outputs from the
                X2 = np.array(
                    X
                )  # minimizing function, find the difference and divide by the perturbation (finding dForce/d change in design variable)
                deltaX = (
                    stepfac * tols[i]
                )  # note: this function uses the tols variable that is computed in dsolve based on the tol input
                X2[i] += deltaX
                Y2, _, _ = eval_func(X2, args)  # here we use the provided eval_func

                J[:, i] = (Y2 - Y) / deltaX  # and append that column to each respective column of the Jacobian matrix

            if N > 1:
                dX = -np.matmul(
                    np.linalg.inv(J), Y - Ytarget
                )  # Take this nth output from the minimizing function and divide it by the jacobian (derivative)
            else:
                if J[0, 0] == 0.0:
                    raise ValueError("dsolve2 found a zero gradient")

                dX = np.array([-(Y[0] - Ytarget[0]) / J[0, 0]])

                if display > 1:
                    print(
                        f" step_func iter {iter} X={X[0]:9.2e}, error={Y[0]-Ytarget[0]:9.2e}, slope={J[0,0]:9.2e}, dX={dX[0]:9.2e}"
                    )

            return dX  # returns dX (step to make)

    # handle bounds
    if len(Xmin) == 0:
        Xmin = np.zeros(N) - np.inf
    elif len(Xmin) == N:
        Xmin = np.array(Xmin, dtype=np.float_)
    else:
        raise TypeError("Xmin must be of same length as X0")

    if len(Xmax) == 0:
        Xmax = np.zeros(N) + np.inf
    elif len(Xmax) == N:
        Xmax = np.array(Xmax, dtype=np.float_)
    else:
        raise TypeError("Xmax must be of same length as X0")

    if len(dX_last) == 0:
        dX_last = np.zeros(N)
    else:
        dX_last = np.array(dX_last, dtype=np.float_)

    if display > 0:
        print(f"Starting dsolve iterations>>>   aiming for Y={Ytarget}")

    for iter in range(maxIter):

        # call evaluation function
        Y, oths, stop = eval_func(X, args)

        # compute error
        err = Y - Ytarget

        if display > 2:
            print(f"  new iteration #{iter} with X={X} and Y={Y}")

        Xs[iter, :] = X
        Es[iter, :] = err

        # stop if commanded by objective function
        if stop:
            break

        # handle tolerances input
        if np.isscalar(tol):
            tols = tol * (np.abs(X) + tol)
        else:
            tols = np.array(tol)

        # check maximum iteration
        if iter == maxIter - 1:
            if display > 0:
                print("Failed to find solution after " + str(iter) + " iterations, with error of " + str(err))

            # looks like things didn't converge, so if N=1 do a linear fit on the last 30% of points to estimate the soln
            if N == 1:

                m, b = np.polyfit(Es[int(0.7 * iter) : iter, 0], Xs[int(0.7 * iter) : iter, 0], 1)
                X = np.array([b])
                Y = np.array([0.0])
                print(f"Using linaer fit to estimate solution at X={b}")

            break

        # >>>> COULD ALSO HAVE AN ITERATION RESTART FUNCTION? >>>
        #  that returns a restart boolean, as well as what values to use to restart things if true. How?

        else:
            dX = step_func(X, args, Y, oths, Ytarget, err, tols, iter, maxIter)

        # Make sure we're not diverging by keeping things from reversing too much.
        # Track the previous step (dX_last) and if the current step reverses too much, stop it part way.
        # Stop it at a plane part way between the current X value and the previous X value (using golden ratio, why not).

        # get the point along the previous step vector where we'll draw the bounding hyperplane (could be a line, plane, or more in higher dimensions)
        Xlim = X - 0.62 * dX_last

        # the equation for the plane we don't want to recross is then sum(X*dX_last) = sum(Xlim*dX_last)
        if np.sum((X + dX) * dX_last) < np.sum(Xlim * dX_last):  # if we cross are going to cross it

            alpha = np.sum((Xlim - X) * dX_last) / np.sum(
                dX * dX_last
            )  # this is how much we need to scale down dX to land on it rather than cross it

            if display > 2:
                print("  limiting oscillation with alpha=" + str(alpha))
                print(f"   dX_last was {dX_last}, dX was going to be {dX}, now it'll be {alpha*dX}")
                print(f"   dX_last was {dX_last/1000}, dX was going to be {dX/1000}, now it'll be {alpha*dX/1000}")

            dX = alpha * dX  # scale down dX

        # also avoid extreme accelerations in the same direction
        for i in range(N):

            if abs(dX_last[i]) > tols[i]:  # only worry about accelerations if the last step was non-negligible

                dX_max = (
                    a_max * dX_last[i]
                )  # set the maximum permissible dx in each direction based an an acceleration limit

                if dX_max == 0.0:  # avoid a divide-by-zero case (if dX[i] was zero to start with)
                    dX[i] = 0.0
                else:
                    a_i = dX[i] / dX_max  # calculate ratio of desired dx to max dx

                    if a_i > 1.0:

                        if display > 2:
                            print(f"    limiting acceleration ({1.0/a_i:6.4f}) for axis {i}")
                            print(f"     dX_last was {dX_last}, dX was going to be {dX}")

                        # dX = dX*a_max/a_i  # scale it down to the maximum value
                        dX[i] = dX[i] / a_i  # scale it down to the maximum value (treat each DOF individually)

                        if display > 2:
                            print(f"     now dX will be {dX}")

        dXlist[iter, :] = dX

        # enforce bounds
        for i in range(N):

            if X[i] + dX[i] < Xmin[i]:
                dX[i] = Xmin[i] - X[i]

            elif X[i] + dX[i] > Xmax[i]:
                dX[i] = Xmax[i] - X[i]

        dXlist2[iter, :] = dX
        # check for convergence
        if all(np.abs(dX) < tols):

            if display > 0:
                print(
                    "Iteration converged after "
                    + str(iter)
                    + " iterations with error of "
                    + str(err)
                    + " and dX of "
                    + str(dX)
                )
                print("Solution X is " + str(X))

                if display > 0:
                    print(
                        "Total run time: {:8.2f} seconds = {:8.2f} minutes".format(
                            (time.time() - start_time), ((time.time() - start_time) / 60)
                        )
                    )

            if any(X == Xmin) or any(X == Xmax):
                success = False
                print("Warning: dsolve ended on a bound.")
            else:
                success = True

            break

        dX_last = 1.0 * dX  # remember this current value

        X = X + dX

    return (
        X,
        Y,
        dict(iter=iter, err=err, dX=dX_last, oths=oths, Xs=Xs, Es=Es, success=success, dXlist=dXlist, dXlist2=dXlist2),
    )
